"""
pyrad.io.config
===============

Functions for reading pyrad config files

.. autosummary::
    :toctree: generated/

    read_config
    get_num_elements
    string_to_datatype
    get_array
    get_struct
    get_array_type
    init_array

"""

import os
import numpy as np
from warnings import warn

_defaults_loc = {
    "ppiImageConfig": {
        "xsize": 10,
        "ysize": 10,
        "dpi": 72,
        "xmin": -50,
        "xmax": 50,
        "ymin": -50,
        "ymax": 50,
    },
    "ppiMapImageConfig": {"xsize": 10, "ysize": 10, "dpi": 72},
    "rhiImageConfig": {
        "xsize": 10,
        "ysize": 5,
        "dpi": 72,
        "xmin": 0,
        "xmax": 50,
        "ymin": 0,
        "ymax": 12,
    },
    "xsecImageConfig": {
        "xsize": 10,
        "ysize": 5,
        "dpi": 72,
        "xmin": 0,
        "xmax": 50,
        "ymin": 0,
        "ymax": 12,
    },
    "gridMapImageConfig": {"xsize": 10, "ysize": 10, "dpi": 72},
    "sunhitsImageConfig": {
        "xsize": 10,
        "ysize": 5,
        "dpi": 72,
        "azmin": -2.0,
        "azmax": 2.0,
        "elmin": -2.0,
        "elmax": 2.0,
        "azres": 0.1,
        "elres": 0.1,
    },
    "spectraImageConfig": {
        "xsize": 10,
        "ysize": 5,
        "dpi": 72,
        "ymin": None,
        "ymax": None,
        "velmin": None,
        "velmax": None,
    },
    "ScanPeriod": 5,
    "IconRunFreq": 3,
    "IconForecasted": 7,
    "RadarName": None,
    "RadarRes": None,
    "NumRadars": 1,
    "TimeTol": 3600.0,
    "ScanList": None,
}

_defaults_main = {
    "lastStateFile": None,
    "datapath": None,
    "satpath": None,
    "iconpath": None,
    "psrpath": None,
    "iqpath": None,
    "colocgatespath": None,
    "excessgatespath": None,
    "dempath": None,
    "smnpath": None,
    "disdropath": None,
    "solarfluxpath": None,
    "selfconsistencypath": None,
    "loadbasepath": None,
    "loadname": None,
    "gecsxbasepath": None,
    "gecsxname": None,
    "metranet_read_lib": "C",
}

DEFAULT_CONFIG = {"loc": _defaults_loc, "main": _defaults_main}


def read_config(fname, cfg=None, defaults=None):
    """
    Read a pyrad config file.

    Parameters
    ----------
    fname : str
        Name of the configuration file to read.

    cfg : dict of dicts, optional
        dictionary of dictionaries containing configuration parameters where
        the new parameters will be placed

    defaults: dict of dicts, optional
        dictionary of dictionaries containing default values. If a key is
        contained in the defaults dict but not in the configuration file,
        the value from the defaults dict will be assigned

    Returns
    -------
    cfg : dict of dicts
        dictionary of dictionaries containing the configuration parameters

    """

    # check if the file can be read
    try:
        cfgfile = open(fname, "r", encoding="utf-8", errors="ignore")
    except Exception:
        raise Exception("ERROR: Could not find|open config file '" + fname + "'")

    # if config dictionary does not exist yet create it
    if cfg is None:
        cfg = dict()

    # if default does not exist, create it
    if defaults is None:
        defaults = dict()

    # read file contents
    fileend = 0
    while fileend == 0:
        # remove leading and trailing whitespace
        line = os.path.expandvars(cfgfile.readline())
        if line:
            line = line.strip()

            # ignore white lines
            if not line:
                continue

            # ignore comments
            if line.startswith("#"):
                continue

            line = line.partition("#")[0]  # Remove comments
            line = line.strip()

            try:
                vals = line.split()

                fieldname = vals[0]
                nvals = len(vals)

                if nvals < 3:
                    raise Exception(
                        "FILE FORMAT ERROR: file: "
                        + fname
                        + ", variable: "
                        + fieldname
                        + ": Wrong number of elements!"
                    )

                valtype = vals[1]
                valuestr = vals[2:nvals]
                nel, isstruct = get_num_elements(valtype, valuestr)

                if nel > 0:
                    if isstruct:
                        pos = cfgfile.tell()
                        fieldvalue, newpos = get_struct(cfgfile, pos, nel, fname)
                        cfgfile.seek(newpos)
                    else:
                        pos = cfgfile.tell()
                        fieldvalue, newpos = get_array(cfgfile, pos, nel, valtype)
                        cfgfile.seek(newpos)
                else:
                    fieldvalue = string_to_datatype(valtype, valuestr)

                cfg.update({fieldname: fieldvalue})
            except BaseException:
                if line != "":
                    warn("Parsing failed after line {:s}".format(line))
                raise
        else:
            fileend = 1

    cfgfile.close()
    # Verify that all keys in default are in newly created config
    cfg = merge_dicts(cfg, defaults)
    return cfg


def get_num_elements(dtype, nelstr):
    """
    Checks if data type is an array or a structure.

    Parameters
    ----------
    dtype : str
        data type specifier

    nelstr : str
        number of elements

    Returns
    -------
    nel : int
        number of elements if type is *ARR or STRUCT. 0 otherwise

    isstruct : bool
        true if the type is STRUCT

    """

    uptype = dtype.upper()
    isstruct = False
    nel = 0
    narr = uptype.count("ARR")
    if uptype == "STRUCT":
        isstruct = True
    if (narr > 0) or (isstruct is True):
        nel = int(nelstr[0])

    return nel, isstruct


def string_to_datatype(dtype, strval):
    """
    Converts a string containing a value into its Python value

    Parameters
    ----------
    dtype : str
        data type specifier

    strval : str
        string value

    Returns
    -------
    val : scalar
        value contained in the string

    """

    uptype = dtype.upper()

    if uptype == "BYTE":
        return int(strval[0])
    elif uptype == "BOOL":
        return bool(strval[0])
    elif uptype == "INT":
        return int(strval[0])
    elif uptype == "LONG":
        return int(strval[0])
    elif uptype == "HEX":
        return int(strval[0])
    elif uptype == "EXP":
        return float(strval[0])
    elif uptype == "FLOAT":
        return float(strval[0])
    elif uptype == "DOUBLE":
        return float(strval[0])
    elif uptype == "STRING":
        # Replace all environement variables
        return os.path.expandvars(str(strval[0]))
    else:
        raise Exception("ERROR: Unexpected data type " + uptype)


def get_array(cfgfile, pos, nel, valtype):
    """
    reads an array in a config file

    Parameters
    ----------
    cfgfile : file object
        config file

    pos : int
        position in file object

    nel : int
        number of elements of the ray

    valtype : str
        type of array

    Returns
    -------
    arr : array
        array values

    newpos : int
        new position in file object

    """

    arr_type = get_array_type(valtype)
    if arr_type == "STRING":
        arr = []
    else:
        arr = init_array(nel, arr_type)

    newpos = pos
    for i in range(nel):
        pos = cfgfile.seek(newpos)
        line = cfgfile.readline()

        # remove leading and trailing whitespace
        line = line.strip()

        line = line.partition("#")[0]  # Remove comments
        line = line.strip()

        vals = line.split()

        value = string_to_datatype(arr_type, vals)
        if arr_type == "STRING":
            arr.append(value)
        else:
            arr[i] = value
        newpos = cfgfile.tell()

    return arr, newpos


def get_struct(cfgfile, pos, nels, fname):
    """
    reads an struct in a config file

    Parameters
    ----------
    cfgfile : file object
        config file
    pos : int
        position in file object
    nel : int
        number of elements of the ray
    fname : str
        config file name

    Returns
    -------
    struct : dict
        dictionary of struct values

    newpos : int
        new position in file object

    """

    struct = dict()
    newpos = pos
    for i in range(nels):
        pos = cfgfile.seek(newpos)
        line = os.path.expandvars(cfgfile.readline())

        # remove leading and trailing whitespace
        line = line.strip()

        line = line.partition("#")[0]  # Remove comments
        line = line.strip()

        try:
            vals = line.split()

            sfieldname = vals[0]
            nvals = len(vals)

            if nvals < 3:
                raise Exception(
                    "FILE FORMAT ERROR: file: "
                    + fname
                    + ", struct variable: "
                    + sfieldname
                    + ": Wrong number of elements!"
                )

            svaltype = vals[1]
            svaluestr = vals[2:nvals]
            nel, isstruct = get_num_elements(svaltype, svaluestr)

            if nel > 0:
                if isstruct:
                    pos = cfgfile.tell()
                    sfieldvalue, newpos = get_struct(cfgfile, pos, nel, fname)
                    cfgfile.seek(newpos)
                else:
                    pos = cfgfile.tell()
                    sfieldvalue, newpos = get_array(cfgfile, pos, nel, svaltype)
                    cfgfile.seek(newpos)
            else:
                newpos = cfgfile.tell()
                sfieldvalue = string_to_datatype(svaltype, svaluestr)

            struct.update({sfieldname: sfieldvalue})
        except BaseException:
            if line != "":
                warn("Parsing failed after line {:s}".format(line))
            raise

    return struct, newpos


def get_array_type(dtype):
    """
    Determines Python array type from the config file array type

    Parameters
    ----------
    dtype : str
        config file data type

    Returns
    -------
    pytype : str
        Python array type

    """

    uptype = dtype.upper()

    if uptype == "BYTARR":
        return "BYTE"
    elif uptype == "INTARR":
        return "INT"
    elif uptype == "LONARR":
        return "LON"
    elif uptype == "HEXARR":
        return "HEX"
    elif uptype == "EXPARR":
        return "EXP"
    elif uptype == "FLTARR":
        return "FLOAT"
    elif uptype == "DBLARR":
        return "DOUBLE"
    elif uptype == "STRARR":
        return "STRING"
    else:
        raise Exception("ERROR: Unexpected data type " + uptype)


def init_array(nel, dtype):
    """
    Initializes a Python array

    Parameters
    ----------
    nel : int
        number of elements in the array
    dtype : str
        config file data type

    Returns
    -------
    pyarr : array
        Python array

    """

    uptype = dtype.upper()

    if uptype == "BYTE":
        return np.empty(nel, dtype="byte")
    elif uptype == "INT":
        return np.empty(nel, dtype="int32")
    elif uptype == "LONG":
        return np.empty(nel, dtype="int64")
    elif uptype == "HEX":
        return np.empty(nel, dtype="H")
    elif uptype == "EXP":
        return np.empty(nel, dtype="float64")
    elif uptype == "FLOAT":
        return np.empty(nel, dtype="float64")
    elif uptype == "DOUBLE":
        return np.empty(nel, dtype="double")
    elif uptype == "STRING":
        return []
    else:
        raise Exception("ERROR: Unexpected array data type " + uptype)


def merge_dicts(ref, defaults):
    """
    Merge two nested dictionaries recursively.

    Parameters:
    -----------
    ref : dict
        The base dictionary to be merged into.
    defaults : dict
        The dictionary whose keys and values are merged into dictionary ref.

    Returns:
    --------
    ref: dict
        The base dictionary after addition of missing keys from defaults
    """
    for key, value in defaults.items():
        if key in ref:
            if isinstance(ref[key], dict) and isinstance(value, dict):
                merge_dicts(ref[key], value)
        else:
            ref[key] = value
    return ref



"""
pyrad.proc.process_timeseries
=============================

Functions to obtain time series of radar data

.. autosummary::
    :toctree: generated/

    process_point_measurement
    process_multiple_points
    process_qvp
    process_rqvp
    process_evp
    process_svp
    process_time_height
    process_ts_along_coord

"""

import datetime
from warnings import warn
import numpy as np
from netCDF4 import num2date

import pyart

from ..io.io_aux import get_datatype_fields, get_fieldname_pyart
from ..io.read_data_sensor import read_coord_sensors


def process_point_measurement(procstatus, dscfg, radar_list=None):
    """
    Obtains the radar data at a point location.

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The data type where we want to extract the point measurement,
            can be any datatype supported by pyrad and available in the data
        agg_method : string. Dataset keyword
            Which aggregation method to use to combine data that is within the
            tolerance in AziTol, EleTol and RngTol
            'nearest' : will only get nearest point to prescribed lon/lat/alt or
                ele/azi/rng
            'nearest_valid' : will only get the nearest valid point to prescribed
                lon/lat/alt or ele/azi/rng (ignore missing values)
            'average' : will average (while ignore missing values), all values
                that fall within the tolerance in AziTol, EleTol and RngTol
            'none' : will not perform any averaging and will get all values that
                fall within the tolerance in AziTol, EleTol and RngTol, each
                with its individual timestamp
            Default is 'nearest'
        latlon : boolean. Dataset keyword
            if True position is obtained from latitude, longitude information,
            otherwise position is obtained from antenna coordinates
            (range, azimuth, elevation).
        truealt : boolean. Dataset keyword
            if True the user input altitude is used to determine the point of
            interest.
            if False use the altitude at a given radar elevation ele over the
            point of interest. Default is False.
        lon : float. Dataset keyword
            the longitude [deg]. Use when latlon is True.
        lat : float. Dataset keyword
            the latitude [deg]. Use when latlon is True.
        alt : float. Dataset keyword
            altitude [m MSL]. Use when latlon is True and truealt is True
        ele : float. Dataset keyword
            radar elevation [deg]. Use when latlon is False or when latlon is
            True and truealt is False
        azi : float. Dataset keyword
            radar azimuth [deg]. Use when latlon is False
        rng : float. Dataset keyword
            range from radar [m]. Use when latlon is False
        AziTol : float. Dataset keyword
            azimuthal tolerance to determine which radar azimuth to use [deg]
        EleTol : float. Dataset keyword
            elevation tolerance to determine which radar elevation to use [deg]
        RngTol : float. Dataset keyword
            range tolerance to determine which radar bin to use [m]
        fill_value : float or None
            If not None masked values are going to be filled by this value

    radar_list : list of Radar objects
          Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the data and metadata at the point of interest
    ind_rad : int
        radar index

    """
    if procstatus == 0:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        break
    field_name = get_fieldname_pyart(datatype)
    ind_rad = int(radarnr[5:8]) - 1

    if procstatus == 2:
        if dscfg["initialized"] == 0:
            return None, None

        # prepare for exit
        new_dataset = {
            "time": dscfg["global_data"]["time"],
            "ref_time": dscfg["global_data"]["ref_time"],
            "datatype": datatype,
            "point_coordinates_WGS84_lon_lat_alt": (
                dscfg["global_data"]["point_coordinates_WGS84_lon_lat_alt"]
            ),
            "antenna_coordinates_az_el_r": (
                dscfg["global_data"]["antenna_coordinates_az_el_r"]
            ),
            "final": True,
        }

        return new_dataset, ind_rad

    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if field_name not in radar.fields:
        warn(
            "Unable to extract point measurement information. " + "Field not available"
        )
        return None, None

    projparams = dict()
    projparams.update({"proj": "pyart_aeqd"})
    projparams.update({"lon_0": radar.longitude["data"]})
    projparams.update({"lat_0": radar.latitude["data"]})

    fill_value = dscfg.get("fill_value", None)
    agg_method = dscfg.get("agg_method", "nearest")
    if agg_method not in ["nearest", "nearest_valid", "average", "none"]:
        warn("Invalid agg_method, use 'nearest', 'nearest_valid', 'average' or 'none'")
        warn("Using 'nearest' instead...")
        agg_method = "nearest"

    if dscfg["latlon"]:
        lon = dscfg["lon"]
        lat = dscfg["lat"]
        x, y = pyart.core.geographic_to_cartesian(lon, lat, projparams)

        if not dscfg["truealt"]:
            ke = 4.0 / 3.0  # constant for effective radius
            a = 6378100.0  # earth radius
            re = a * ke  # effective radius

            elrad = dscfg["ele"] * np.pi / 180.0
            r_ground = np.sqrt(x**2.0 + y**2.0)
            r = r_ground / np.cos(elrad)
            alt = (
                radar.altitude["data"]
                + np.sqrt(r**2.0 + re**2.0 + 2.0 * r * re * np.sin(elrad))
                - re
            )
            alt = alt[0]
        else:
            alt = dscfg["alt"]

        r, az, el = pyart.core.cartesian_to_antenna(x, y, alt - radar.altitude["data"])
        r = r[0]
        az = az[0]
        el = el[0]
    else:
        r = dscfg["rng"]
        az = dscfg["azi"]
        el = dscfg["ele"]

        x, y, alt = pyart.core.antenna_to_cartesian(r / 1000.0, az, el)
        lon, lat = pyart.core.cartesian_to_geographic(x, y, projparams)
        lon = lon[0]
        lat = lat[0]

    d_az = np.abs(radar.azimuth["data"] - az)
    if np.min(d_az) > dscfg["AziTol"]:
        warn(
            " No radar bin found for point (az, el, r):("
            + str(az)
            + ", "
            + str(el)
            + ", "
            + str(r)
            + "). Minimum distance to radar azimuth "
            + str(np.min(d_az))
            + " larger than tolerance"
        )
        return None, None

    d_el = np.abs(radar.elevation["data"] - el)
    if np.min(d_el) > dscfg["EleTol"]:
        warn(
            " No radar bin found for point (az, el, r):("
            + str(az)
            + ", "
            + str(el)
            + ", "
            + str(r)
            + "). Minimum distance to radar elevation "
            + str(np.min(d_el))
            + " larger than tolerance"
        )
        return None, None

    d_r = np.abs(radar.range["data"] - r)
    if np.min(d_r) > dscfg["RngTol"]:
        warn(
            " No radar bin found for point (az, el, r):("
            + str(az)
            + ", "
            + str(el)
            + ", "
            + str(r)
            + "). Minimum distance to radar range bin "
            + str(np.min(d_r))
            + " larger than tolerance"
        )
        return None, None

    if agg_method == "nearest":
        # Get closest gate
        ind_ray = np.argmin(
            np.abs(radar.azimuth["data"] - az) + np.abs(radar.elevation["data"] - el)
        )
        ind_r = np.argmin(np.abs(radar.range["data"] - r))
        val = radar.fields[field_name]["data"][ind_ray, ind_r]
    else:
        # Get all gates that fall within tolerance
        ind_ray = np.where(
            np.logical_and(d_az <= dscfg["AziTol"], d_el <= dscfg["EleTol"])
        )[0]
        ind_r = np.where(np.abs(radar.range["data"] - r) < dscfg["RngTol"])[0]

        if agg_method == "average":
            val = np.nanmean(radar.fields[field_name]["data"].data[ind_ray][:, ind_r])
        elif agg_method == "none":
            val = radar.fields[field_name]["data"].data[ind_ray, ind_r]
        elif agg_method == "nearest_valid":
            ind_clo_ray = np.argmin(
                np.abs(radar.azimuth["data"] - az)
                + np.abs(radar.elevation["data"] - el)
            )
            ind_clo_r = np.argmin(np.abs(radar.range["data"] - r))

            # We get x,y,z position of closest gate to objective
            clo_x = radar.gate_x["data"][ind_clo_ray, ind_clo_r]
            clo_y = radar.gate_y["data"][ind_clo_ray, ind_clo_r]
            clo_z = radar.gate_z["data"][ind_clo_ray, ind_clo_r]

            # We get the distance from the closest gate to every gate within tolerance
            dist_x = np.abs(radar.gate_x["data"][ind_ray][:, ind_r] - clo_x)
            dist_y = np.abs(radar.gate_y["data"][ind_ray][:, ind_r] - clo_y)
            dist_z = np.abs(radar.gate_z["data"][ind_ray][:, ind_r] - clo_z)

            dist_to_clo = np.sqrt(dist_x**2 + dist_y**2 + dist_z**2)
            val = radar.fields[field_name]["data"][ind_ray][:, ind_r]
            # Assign inf distance to invalid gates
            dist_to_clo[val.mask] = np.inf

            # Find closest gate which is valid
            clo_valid_idx = np.unravel_index(np.argmin(dist_to_clo), dist_to_clo.shape)
            val = val[clo_valid_idx]
            ind_ray = ind_ray[clo_valid_idx[0]]
            ind_r = ind_r[clo_valid_idx[1]]

    ant_coord = np.empty((3, ind_ray.size), np.float32)
    ant_coord[0, :] = radar.azimuth["data"][ind_ray]
    ant_coord[1, :] = radar.elevation["data"][ind_ray]
    ant_coord[2, :] = np.zeros(ind_ray.size) + radar.range["data"][ind_r]

    if fill_value is not None:
        try:
            val = val.filled(fill_value)
        except AttributeError:
            # val is not masked array
            pass

    time = num2date(
        radar.time["data"][ind_ray], radar.time["units"], radar.time["calendar"]
    )

    # initialize dataset
    if dscfg["initialized"] == 0:
        poi = {
            "point_coordinates_WGS84_lon_lat_alt": [lon, lat, alt],
            "antenna_coordinates_az_el_r": [az, el, r],
            "time": time,
            "ref_time": dscfg["timeinfo"],
        }
        dscfg["global_data"] = poi
        dscfg["initialized"] = 1

    dscfg["global_data"]["ref_time"] = dscfg["timeinfo"]

    # prepare for exit
    new_dataset = dict()
    new_dataset.update({"value": val})
    new_dataset.update({"datatype": datatype})
    new_dataset.update({"time": time})
    new_dataset.update({"ref_time": dscfg["timeinfo"]})
    new_dataset.update({"point_coordinates_WGS84_lon_lat_alt": [lon, lat, alt]})
    new_dataset.update({"antenna_coordinates_az_el_r": [az, el, r]})
    new_dataset.update({"used_antenna_coordinates_az_el_r": ant_coord})
    new_dataset.update({"final": False})
    return new_dataset, ind_rad


def process_multiple_points(procstatus, dscfg, radar_list=None):
    """
    Obtains the radar data at multiple points. The points are defined in a file

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The data type where we want to extract the point measurement,
            can be any datatype supported by pyrad and available in the data
        truealt : boolean. Dataset keyword
            if True the user input altitude is used to determine the point of
            interest.
            if False use the altitude at a given radar elevation ele over the
            point of interest. Default is False.
        coord_fname : string
            File name containing the points coordinates
        alt_points : float. Dataset keyword
            altitude [m MSL]. Use when latlon is True.
        ele_points : float. Dataset keyword
            radar elevation [deg]. Use when latlon is False or when latlon is
            True and truealt is False
        AziTol : float. Dataset keyword
            azimuthal tolerance to determine which radar azimuth to use [deg]
        EleTol : float. Dataset keyword
            elevation tolerance to determine which radar elevation to use [deg]
        RngTol : float. Dataset keyword
            range tolerance to determine which radar bin to use [m]

    radar_list : list of Radar objects
          Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the data and metadata at the point of interest
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg["datatype"]:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        break
    field_name = get_fieldname_pyart(datatype)
    ind_rad = int(radarnr[5:8]) - 1

    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn("ERROR: No valid radar")
        return None, None
    radar = radar_list[ind_rad]

    if field_name not in radar.fields:
        warn(
            "Unable to extract point measurement information. " + "Field not available"
        )
        return None, None

    # default parameters
    truealt = dscfg.get("truealt", False)
    ele_points = dscfg.get("ele_points", 1.0)
    alt_points = dscfg.get("alt_points", 0.0)
    AziTol = dscfg.get("AziTol", 0.25)
    EleTol = dscfg.get("EleTol", 1.0)
    RngTol = dscfg.get("RngTol", 120.0)

    projparams = dict()
    projparams.update({"proj": "pyart_aeqd"})
    projparams.update({"lon_0": radar.longitude["data"]})
    projparams.update({"lat_0": radar.latitude["data"]})

    lat, lon, point_id = read_coord_sensors(dscfg["coord_fname"])
    x, y = pyart.core.geographic_to_cartesian(lon, lat, projparams)
    npoints = lon.size

    if not truealt:
        ke = 4.0 / 3.0  # constant for effective radius
        a = 6378100.0  # earth radius
        re = a * ke  # effective radius

        elrad = ele_points * np.pi / 180.0
        r_ground = np.sqrt(x**2.0 + y**2.0)
        r = r_ground / np.cos(elrad)
        alt = (
            radar.altitude["data"]
            + np.sqrt(r**2.0 + re**2.0 + 2.0 * r * re * np.sin(elrad))
            - re
        )
    else:
        alt = alt_points + np.zeros(npoints)

    r, az, el = pyart.core.cartesian_to_antenna(x, y, alt - radar.altitude["data"])

    val = np.ma.masked_all(npoints)
    time = np.ma.masked_all(npoints, dtype=datetime.datetime)
    used_lon = np.ma.masked_all(npoints)
    used_lat = np.ma.masked_all(npoints)
    used_alt = np.ma.masked_all(npoints)
    for ind in range(npoints):
        d_az = np.abs(radar.azimuth["data"] - az[ind])
        if np.min(d_az) > AziTol:
            warn(
                " No radar bin found for point (az, el, r):("
                + str(az[ind])
                + ", "
                + str(el[ind])
                + ", "
                + str(r[ind])
                + "). Minimum distance to radar azimuth "
                + str(d_az)
                + " larger than tolerance"
            )
            continue

        d_el = np.abs(radar.elevation["data"] - el[ind])
        if np.min(d_el) > EleTol:
            warn(
                " No radar bin found for point (az, el, r):("
                + str(az[ind])
                + ", "
                + str(el[ind])
                + ", "
                + str(r[ind])
                + "). Minimum distance to radar elevation "
                + str(d_el)
                + " larger than tolerance"
            )
            continue

        d_r = np.abs(radar.range["data"] - r[ind])
        if np.min(d_r) > RngTol:
            warn(
                " No radar bin found for point (az, el, r):("
                + str(az[ind])
                + ", "
                + str(el[ind])
                + ", "
                + str(r[ind])
                + "). Minimum distance to radar range bin "
                + str(d_r)
                + " larger than tolerance"
            )
            continue

        ind_ray = np.argmin(
            np.abs(radar.azimuth["data"] - az[ind])
            + np.abs(radar.elevation["data"] - el[ind])
        )
        ind_r = np.argmin(np.abs(radar.range["data"] - r[ind]))

        used_lon[ind] = radar.gate_longitude["data"][ind_ray, ind_r]
        used_lat[ind] = radar.gate_latitude["data"][ind_ray, ind_r]
        used_alt[ind] = radar.gate_altitude["data"][ind_ray, ind_r]

        val[ind] = radar.fields[field_name]["data"].data[ind_ray, ind_r]
        time[ind] = num2date(
            radar.time["data"][ind_ray], radar.time["units"], radar.time["calendar"]
        )

    # prepare for exit
    new_dataset = dict()
    new_dataset.update({"value": val})
    new_dataset.update({"datatype": datatype})
    new_dataset.update({"time": time})
    new_dataset.update({"ref_time": dscfg["timeinfo"]})
    new_dataset.update({"point_coordinates_WGS84_lon": lon})
    new_dataset.update({"point_coordinates_WGS84_lat": lat})
    new_dataset.update({"point_coordinates_WGS84_alt": alt})
    new_dataset.update({"antenna_coordinates_az": az})
    new_dataset.update({"antenna_coordinates_el": el})
    new_dataset.update({"antenna_coordinates_r": r})
    new_dataset.update({"used_point_coordinates_WGS84_lon": used_lon})
    new_dataset.update({"used_point_coordinates_WGS84_lat": used_lat})
    new_dataset.update({"used_point_coordinates_WGS84_alt": used_alt})
    new_dataset.update({"point_id": point_id})
    new_dataset.update({"final": False})

    return new_dataset, ind_rad


def process_qvp(procstatus, dscfg, radar_list=None):
    """
    Computes quasi vertical profiles, by averaging over height levels
    PPI data.

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The data type where we want to extract the point measurement,
            can be any datatype supported by pyrad and available in the data
        angle : int or float
            If the radar object contains a PPI volume, the sweep number to
            use, if it contains an RHI volume the elevation angle.
            Default 0.
        ang_tol : float
            If the radar object contains an RHI volume, the tolerance in the
            elevation angle for the conversion into PPI
        hmax : float
            The maximum height to plot [m]. Default 10000.
        hres : float
            The height resolution [m]. Default 50
        avg_type : str
            The type of averaging to perform. Can be either "mean" or "median"
            Default "mean"
        nvalid_min : int
            Minimum number of valid points to accept average. Default 30.
        interp_kind : str
            type of interpolation when projecting to vertical grid: 'none',
            or 'nearest', etc. Default 'none'
            'none' will select from all data points within the regular grid
            height bin the closest to the center of the bin.
            'nearest' will select the closest data point to the center of the
            height bin regardless if it is within the height bin or not.
            Data points can be masked values
            If another type of interpolation is selected masked values will be
            eliminated from the data points before the interpolation

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the QVP and a keyword stating whether the
        processing has finished or not.
    ind_rad : int
        radar index

    Reference
    ---------
    Ryzhkov A., Zhang P., Reeves H., Kumjian M., Tschallener T., Trömel S.,
    Simmer C. 2016: Quasi-Vertical Profiles: A New Way to Look at Polarimetric
    Radar Data. JTECH vol. 33 pp 551-562

    """
    if procstatus == 0:
        return None, None

    if procstatus == 1:
        field_names = []
        for datatypedescr in dscfg["datatype"]:
            radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
            field_names.append(get_fieldname_pyart(datatype))

        ind_rad = int(radarnr[5:8]) - 1

        if (radar_list is None) or (radar_list[ind_rad] is None):
            warn("ERROR: No valid radar")
            return None, None

        radar = radar_list[ind_rad]

        # default parameters
        angle = dscfg.get("angle", 0)
        ang_tol = dscfg.get("ang_tol", 1.0)
        hmax = dscfg.get("hmax", 10000.0)
        hres = dscfg.get("hres", 50.0)
        avg_type = dscfg.get("avg_type", "mean")
        nvalid_min = dscfg.get("nvalid_min", 30)
        interp_kind = dscfg.get("interp_kind", "none")

        # initialize dataset
        if not dscfg["initialized"]:
            qvp = pyart.retrieve.compute_qvp(
                radar,
                field_names,
                ref_time=dscfg["timeinfo"],
                angle=angle,
                ang_tol=ang_tol,
                hmax=hmax,
                hres=hres,
                avg_type=avg_type,
                nvalid_min=nvalid_min,
                interp_kind=interp_kind,
                qvp=None,
            )

            if qvp is None:
                warn("Unable to compute QVP")
                return None, None

            global_dict = dict()
            global_dict.update({"start_time": dscfg["timeinfo"]})
            global_dict.update({"radar_out": qvp})
            dscfg["global_data"] = global_dict
            dscfg["initialized"] = 1
        else:
            qvp = pyart.retrieve.compute_qvp(
                radar,
                field_names,
                ref_time=dscfg["timeinfo"],
                angle=angle,
                ang_tol=ang_tol,
                hmax=hmax,
                hres=hres,
                avg_type=avg_type,
                nvalid_min=nvalid_min,
                interp_kind=interp_kind,
                qvp=dscfg["global_data"]["radar_out"],
            )

            if qvp is None:
                warn("Unable to compute QVP")
                return None, None

            dscfg["global_data"]["radar_out"] = qvp

        new_dataset = dict()
        new_dataset.update({"radar_out": qvp})
        new_dataset.update({"radar_type": "temporal"})
        new_dataset.update({"start_time": dscfg["global_data"]["start_time"]})

        return new_dataset, ind_rad

    if procstatus == 2:
        if not dscfg["initialized"]:
            return None, None

        for datatypedescr in dscfg["datatype"]:
            radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
            break

        ind_rad = int(radarnr[5:8]) - 1

        qvp = dscfg["global_data"]["radar_out"]

        new_dataset = dict()
        new_dataset.update({"radar_out": qvp})
        new_dataset.update({"radar_type": "final"})
        new_dataset.update({"start_time": dscfg["global_data"]["start_time"]})

        return new_dataset, ind_rad


def process_rqvp(procstatus, dscfg, radar_list=None):
    """
    Computes range defined quasi vertical profiles, by averaging over height
    levels PPI data.

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The data type where we want to extract the point measurement,
            can be any datatype supported by pyrad and available in the data
        hmax : float
            The maximum height to plot [m]. Default 10000.
        hres : float
            The height resolution [m]. Default 2.
        avg_type : str
            The type of averaging to perform. Can be either "mean" or "median"
            Default "mean"
        nvalid_min : int
            Minimum number of valid points to accept average. Default 30.
        interp_kind : str
            type of interpolation when projecting to vertical grid: 'none',
            or 'nearest', etc. Default 'nearest'
            'none' will select from all data points within the regular grid
            height bin the closest to the center of the bin.
            'nearest' will select the closest data point to the center of the
            height bin regardless if it is within the height bin or not.
            Data points can be masked values
            If another type of interpolation is selected masked values will be
            eliminated from the data points before the interpolation
        rmax : float
            ground range up to which the data is intended for use [m].
            Default 50000.
        weight_power : float
            Power p of the weighting function 1/abs(grng-(rmax-1))**p given to
            the data outside the desired range. -1 will set the weight to 0.
            Default 2.

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the QVP and a keyword stating whether the
        processing has finished or not.
    ind_rad : int
        radar index

    Reference
    ---------
    Tobin D.M., Kumjian M.R. 2017: Polarimetric Radar and Surface-Based
    Precipitation-Type Observations of ice Pellet to Freezing Rain
    Transitions. Weather and Forecasting vol. 32 pp 2065-2082

    """
    if procstatus == 0:
        return None, None

    if procstatus == 1:
        field_names = []
        for datatypedescr in dscfg["datatype"]:
            radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
            field_names.append(get_fieldname_pyart(datatype))

        ind_rad = int(radarnr[5:8]) - 1

        if (radar_list is None) or (radar_list[ind_rad] is None):
            warn("ERROR: No valid radar")
            return None, None

        radar = radar_list[ind_rad]

        # default parameters
        hmax = dscfg.get("hmax", 10000.0)
        hres = dscfg.get("hres", 2.0)
        avg_type = dscfg.get("avg_type", "mean")
        nvalid_min = dscfg.get("nvalid_min", 30)
        interp_kind = dscfg.get("interp_kind", "nearest")
        rmax = dscfg.get("rmax", 50000.0)
        weight_power = dscfg.get("weight_power", 2.0)

        # initialize dataset
        if not dscfg["initialized"]:
            qvp = pyart.retrieve.compute_rqvp(
                radar,
                field_names,
                ref_time=dscfg["timeinfo"],
                hmax=hmax,
                hres=hres,
                avg_type=avg_type,
                nvalid_min=nvalid_min,
                interp_kind=interp_kind,
                rmax=rmax,
                weight_power=weight_power,
                qvp=None,
            )

            if qvp is None:
                warn("Unable to compute QVP")
                return None, None

            global_dict = dict()
            global_dict.update({"start_time": dscfg["timeinfo"]})
            global_dict.update({"radar_out": qvp})
            dscfg["global_data"] = global_dict
            dscfg["initialized"] = 1
        else:
            qvp = pyart.retrieve.compute_rqvp(
                radar,
                field_names,
                ref_time=dscfg["timeinfo"],
                hmax=hmax,
                hres=hres,
                avg_type=avg_type,
                nvalid_min=nvalid_min,
                interp_kind=interp_kind,
                rmax=rmax,
                weight_power=weight_power,
                qvp=dscfg["global_data"]["radar_out"],
            )

            if qvp is None:
                warn("Unable to compute QVP")
                return None, None

            dscfg["global_data"]["radar_out"] = qvp

        new_dataset = dict()
        new_dataset.update({"radar_out": qvp})
        new_dataset.update({"radar_type": "temporal"})
        new_dataset.update({"start_time": dscfg["global_data"]["start_time"]})

        return new_dataset, ind_rad

    if procstatus == 2:
        if not dscfg["initialized"]:
            return None, None

        for datatypedescr in dscfg["datatype"]:
            radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
            break

        ind_rad = int(radarnr[5:8]) - 1

        qvp = dscfg["global_data"]["radar_out"]

        new_dataset = dict()
        new_dataset.update({"radar_out": qvp})
        new_dataset.update({"radar_type": "final"})
        new_dataset.update({"start_time": dscfg["global_data"]["start_time"]})

        return new_dataset, ind_rad


def process_evp(procstatus, dscfg, radar_list=None):
    """
    Computes enhanced vertical profiles, by averaging over height levels
    PPI data.

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The data type where we want to extract the point measurement,
            can be any datatype supported by pyrad and available in the data
        lat, lon : float
            latitude and longitude of the point of interest [deg]
        latlon_tol : float
            tolerance in latitude and longitude in deg. Default 0.0005
        delta_rng, delta_azi : float
            maximum range distance [m] and azimuth distance [degree] from the
            central point of the evp containing data to average. Default 5000.
            and 10.
        hmax : float
            The maximum height to plot [m]. Default 10000.
        hres : float
            The height resolution [m]. Default 250.
        avg_type : str
            The type of averaging to perform. Can be either "mean" or "median"
            Default "mean"
        nvalid_min : int
            Minimum number of valid points to consider the data valid when
            performing the averaging. Default 1
        interp_kind : str
            type of interpolation when projecting to vertical grid: 'none',
            or 'nearest', etc. Default 'none'.
            'none' will select from all data points within the regular grid
            height bin the closest to the center of the bin.
            'nearest' will select the closest data point to the center of the
            height bin regardless if it is within the height bin or not.
            Data points can be masked values
            If another type of interpolation is selected masked values will be
            eliminated from the data points before the interpolation

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the EVP and a keyword stating whether the
        processing has finished or not.
    ind_rad : int
        radar index

    Reference
    ---------
    Kaltenboeck R., Ryzhkov A. 2016: A freezing rain storm explored with a
    C-band polarimetric weather radar using the QVP methodology.
    Meteorologische Zeitschrift vol. 26 pp 207-222

    """
    if procstatus == 0:
        return None, None

    if procstatus == 1:
        field_names = []
        for datatypedescr in dscfg["datatype"]:
            radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
            field_names.append(get_fieldname_pyart(datatype))

        ind_rad = int(radarnr[5:8]) - 1

        if (radar_list is None) or (radar_list[ind_rad] is None):
            warn("ERROR: No valid radar")
            return None, None

        radar = radar_list[ind_rad]

        # default parameters
        lon = dscfg["lon"]
        lat = dscfg["lat"]
        latlon_tol = dscfg.get("latlon_tol", 0.0005)
        delta_rng = dscfg.get("delta_rng", 15000.0)
        delta_azi = dscfg.get("delta_azi", 10.0)
        hmax = dscfg.get("hmax", 10000.0)
        hres = dscfg.get("hres", 250.0)
        avg_type = dscfg.get("avg_type", "mean")
        nvalid_min = dscfg.get("nvalid_min", 1)
        interp_kind = dscfg.get("interp_kind", "none")

        # initialize dataset
        if not dscfg["initialized"]:
            qvp = pyart.retrieve.compute_evp(
                radar,
                field_names,
                lon,
                lat,
                ref_time=dscfg["timeinfo"],
                latlon_tol=latlon_tol,
                delta_rng=delta_rng,
                delta_azi=delta_azi,
                hmax=hmax,
                hres=hres,
                avg_type=avg_type,
                nvalid_min=nvalid_min,
                interp_kind=interp_kind,
                qvp=None,
            )

            if qvp is None:
                warn("Unable to compute QVP")
                return None, None

            global_dict = dict()
            global_dict.update({"start_time": dscfg["timeinfo"]})
            global_dict.update({"radar_out": qvp})
            dscfg["global_data"] = global_dict
            dscfg["initialized"] = 1
        else:
            qvp = pyart.retrieve.compute_evp(
                radar,
                field_names,
                lon,
                lat,
                ref_time=dscfg["timeinfo"],
                latlon_tol=latlon_tol,
                delta_rng=delta_rng,
                delta_azi=delta_azi,
                hmax=hmax,
                hres=hres,
                avg_type=avg_type,
                nvalid_min=nvalid_min,
                interp_kind=interp_kind,
                qvp=dscfg["global_data"]["radar_out"],
            )

            if qvp is None:
                warn("Unable to compute QVP")
                return None, None

        dscfg["global_data"]["radar_out"] = qvp

        new_dataset = dict()
        new_dataset.update({"radar_out": qvp})
        new_dataset.update({"radar_type": "temporal"})
        new_dataset.update({"start_time": dscfg["global_data"]["start_time"]})

        return new_dataset, ind_rad

    if procstatus == 2:
        if not dscfg["initialized"]:
            return None, None

        for datatypedescr in dscfg["datatype"]:
            radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
            break

        ind_rad = int(radarnr[5:8]) - 1

        qvp = dscfg["global_data"]["radar_out"]

        new_dataset = dict()
        new_dataset.update({"radar_out": qvp})
        new_dataset.update({"radar_type": "final"})
        new_dataset.update({"start_time": dscfg["global_data"]["start_time"]})

        return new_dataset, ind_rad


def process_svp(procstatus, dscfg, radar_list=None):
    """
    Computes slanted vertical profiles, by averaging over height levels
    PPI data.

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The data type where we want to extract the point measurement,
            can be any datatype supported by pyrad and available in the data
        angle : int or float
            If the radar object contains a PPI volume, the sweep number to
            use, if it contains an RHI volume the elevation angle.
            Default 0.
        ang_tol : float
            If the radar object contains an RHI volume, the tolerance in the
            elevation angle for the conversion into PPI. Default 1.
        lat, lon : float
            latitude and longitude of the point of interest [deg]
        latlon_tol : float
            tolerance in latitude and longitude in deg. Default 0.0005
        delta_rng, delta_azi : float
            maximum range distance [m] and azimuth distance [degree] from the
            central point of the svp containing data to average. Default 5000.
            and 10.
        hmax : float
            The maximum height to plot [m]. Default 10000.
        hres : float
            The height resolution [m]. Default 250.
        avg_type : str
            The type of averaging to perform. Can be either "mean" or "median"
            Default "mean"
        nvalid_min : int
            Minimum number of valid points to consider the data valid when
            performing the averaging. Default 1
        interp_kind : str
            type of interpolation when projecting to vertical grid: 'none',
            or 'nearest', etc. Default 'none'
            'none' will select from all data points within the regular grid
            height bin the closest to the center of the bin.
            'nearest' will select the closest data point to the center of the
            height bin regardless if it is within the height bin or not.
            Data points can be masked values
            If another type of interpolation is selected masked values will be
            eliminated from the data points before the interpolation

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the svp and a keyword stating whether the
        processing has finished or not.
    ind_rad : int
        radar index

    Reference
    ---------
    Bukovcic P., Zrnic D., Zhang G. 2017: Winter Precipitation Liquid-Ice
    Phase Transitions Revealed with Polarimetric Radar and 2DVD Observations
    in Central Oklahoma. JTECH vol. 56 pp 1345-1363

    """
    if procstatus == 0:
        return None, None

    if procstatus == 1:
        field_names = []
        for datatypedescr in dscfg["datatype"]:
            radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
            field_names.append(get_fieldname_pyart(datatype))

        ind_rad = int(radarnr[5:8]) - 1

        if (radar_list is None) or (radar_list[ind_rad] is None):
            warn("ERROR: No valid radar")
            return None, None

        radar = radar_list[ind_rad]

        # default parameters
        angle = dscfg.get("angle", 0)
        ang_tol = dscfg.get("ang_tol", 1.0)
        lon = dscfg["lon"]
        lat = dscfg["lat"]
        latlon_tol = dscfg.get("latlon_tol", 0.0005)
        delta_rng = dscfg.get("delta_rng", 30000.0)
        delta_azi = dscfg.get("delta_azi", 10.0)
        hmax = dscfg.get("hmax", 10000.0)
        hres = dscfg.get("hres", 250.0)
        avg_type = dscfg.get("avg_type", "mean")
        nvalid_min = dscfg.get("nvalid_min", 1)
        interp_kind = dscfg.get("interp_kind", "none")

        # initialize dataset
        if not dscfg["initialized"]:
            qvp = pyart.retrieve.compute_svp(
                radar,
                field_names,
                lon,
                lat,
                angle,
                ref_time=dscfg["timeinfo"],
                ang_tol=ang_tol,
                latlon_tol=latlon_tol,
                delta_rng=delta_rng,
                delta_azi=delta_azi,
                hmax=hmax,
                hres=hres,
                avg_type=avg_type,
                nvalid_min=nvalid_min,
                interp_kind=interp_kind,
                qvp=None,
            )

            if qvp is None:
                warn("Unable to compute QVP")
                return None, None

            global_dict = dict()
            global_dict.update({"start_time": dscfg["timeinfo"]})
            global_dict.update({"radar_out": qvp})
            dscfg["global_data"] = global_dict
            dscfg["initialized"] = 1
        else:
            qvp = pyart.retrieve.compute_svp(
                radar,
                field_names,
                lon,
                lat,
                angle,
                ref_time=dscfg["timeinfo"],
                ang_tol=ang_tol,
                latlon_tol=latlon_tol,
                delta_rng=delta_rng,
                delta_azi=delta_azi,
                hmax=hmax,
                hres=hres,
                avg_type=avg_type,
                nvalid_min=nvalid_min,
                interp_kind=interp_kind,
                qvp=dscfg["global_data"]["radar_out"],
            )

            if qvp is None:
                warn("Unable to compute QVP")
                return None, None

        dscfg["global_data"]["radar_out"] = qvp

        new_dataset = dict()
        new_dataset.update({"radar_out": qvp})
        new_dataset.update({"radar_type": "temporal"})
        new_dataset.update({"start_time": dscfg["global_data"]["start_time"]})

        return new_dataset, ind_rad

    if procstatus == 2:
        if not dscfg["initialized"]:
            return None, None

        for datatypedescr in dscfg["datatype"]:
            radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
            break

        ind_rad = int(radarnr[5:8]) - 1

        svp = dscfg["global_data"]["radar_out"]

        new_dataset = dict()
        new_dataset.update({"radar_out": svp})
        new_dataset.update({"radar_type": "final"})
        new_dataset.update({"start_time": dscfg["global_data"]["start_time"]})

        return new_dataset, ind_rad


def process_time_height(procstatus, dscfg, radar_list=None):
    """
    Produces time height radar objects at a point of interest defined by
    latitude and longitude. A time-height contains the evolution
    of the vertical structure of radar measurements above the location
    of interest.

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The data type where we want to extract the point measurement,
            can be any datatype supported by pyrad and available in the data
        lat, lon : float
            latitude and longitude of the point of interest [deg]
        latlon_tol : float
            tolerance in latitude and longitude in deg. Default 0.0005
        hmax : float
            The maximum height to plot [m]. Default 10000.
        hres : float
            The height resolution [m]. Default 50
        interp_kind : str
            type of interpolation when projecting to vertical grid: 'none',
            or 'nearest', etc. Default 'none'
            'none' will select from all data points within the regular grid
            height bin the closest to the center of the bin.
            'nearest' will select the closest data point to the center of the
            height bin regardless if it is within the height bin or not.
            Data points can be masked values
            If another type of interpolation is selected masked values will be
            eliminated from the data points before the interpolation

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the QVP and a keyword stating whether the
        processing has finished or not.
    ind_rad : int
        radar index

    """
    if procstatus == 0:
        return None, None

    if procstatus == 1:
        field_names = []
        for datatypedescr in dscfg["datatype"]:
            radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
            field_names.append(get_fieldname_pyart(datatype))

        ind_rad = int(radarnr[5:8]) - 1

        if (radar_list is None) or (radar_list[ind_rad] is None):
            warn("ERROR: No valid radar")
            return None, None

        radar = radar_list[ind_rad]

        # default parameters
        lon = dscfg["lon"]
        lat = dscfg["lat"]
        latlon_tol = dscfg.get("latlon_tol", 0.0005)
        hmax = dscfg.get("hmax", 10000.0)
        hres = dscfg.get("hres", 50.0)
        interp_kind = dscfg.get("interp_kind", "none")

        # initialize dataset
        if not dscfg["initialized"]:
            qvp = pyart.retrieve.compute_vp(
                radar,
                field_names,
                lon,
                lat,
                ref_time=dscfg["timeinfo"],
                latlon_tol=latlon_tol,
                hmax=hmax,
                hres=hres,
                interp_kind=interp_kind,
                qvp=None,
            )

            if qvp is None:
                warn("Unable to compute QVP")
                return None, None

            global_dict = dict()
            global_dict.update({"start_time": dscfg["timeinfo"]})
            global_dict.update({"radar_out": qvp})
            dscfg["global_data"] = global_dict
            dscfg["initialized"] = 1
        else:
            qvp = pyart.retrieve.compute_vp(
                radar,
                field_names,
                lon,
                lat,
                ref_time=dscfg["timeinfo"],
                latlon_tol=latlon_tol,
                hmax=hmax,
                hres=hres,
                interp_kind=interp_kind,
                qvp=dscfg["global_data"]["radar_out"],
            )

            if qvp is None:
                warn("Unable to compute QVP")
                return None, None

        dscfg["global_data"]["radar_out"] = qvp

        new_dataset = dict()
        new_dataset.update({"radar_out": qvp})
        new_dataset.update({"radar_type": "temporal"})
        new_dataset.update({"start_time": dscfg["global_data"]["start_time"]})

        return new_dataset, ind_rad

    if procstatus == 2:
        if not dscfg["initialized"]:
            return None, None

        for datatypedescr in dscfg["datatype"]:
            radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
            break

        ind_rad = int(radarnr[5:8]) - 1

        qvp = dscfg["global_data"]["radar_out"]

        new_dataset = dict()
        new_dataset.update({"radar_out": qvp})
        new_dataset.update({"radar_type": "final"})
        new_dataset.update({"start_time": dscfg["global_data"]["start_time"]})

        return new_dataset, ind_rad


def process_ts_along_coord(procstatus, dscfg, radar_list=None):
    """
    Produces time series along a particular antenna coordinate

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : string. Dataset keyword
            The data type where we want to extract the time series,
            can be any datatype supported by pyrad and available in the data
        mode : str
            coordinate to extract data along. Can be ALONG_AZI, ALONG_ELE or
            ALONG_RNG
        fixed_range, fixed_azimuth, fixed_elevation : float
            The fixed range [m], azimuth [deg] or elevation [deg] to extract.
            In each mode two of these parameters have to be defined. If they
            are not defined they default to 0.
        ang_tol, rng_tol : float
            The angle tolerance [deg] and range tolerance [m] around the fixed
            range or azimuth/elevation
        value_start, value_stop : float
            The minimum and maximum value at which the data along a coordinate
            start and stop

    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the data and a keyword stating whether the
        processing has finished or not.
    ind_rad : int
        radar index

    """
    if procstatus == 0:
        return None, None

    if procstatus == 1:
        radarnr, _, datatype, _, _ = get_datatype_fields(dscfg["datatype"][0])
        field_names = []
        for datatypedescr in dscfg["datatype"]:
            radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
            field_names.append(get_fieldname_pyart(datatype))

        ind_rad = int(radarnr[5:8]) - 1

        if (radar_list is None) or (radar_list[ind_rad] is None):
            warn("ERROR: No valid radar")
            return None, None

        radar = radar_list[ind_rad]

        mode = dscfg.get("mode", "ALONG_AZI")
        if mode == "ALONG_RNG":
            value_start = dscfg.get("value_start", 0.0)
            value_stop = dscfg.get("value_stop", radar.range["data"][-1])
            ang_tol = dscfg.get("AngTol", 1.0)
            rng_tol = dscfg.get("RngTol", 50.0)
            fixed_range = dscfg.get("fixed_range", None)
            fixed_azimuth = dscfg.get("fixed_azimuth", 0.0)
            fixed_elevation = dscfg.get("fixed_elevation", 0.0)
        elif mode == "ALONG_AZI":
            value_start = dscfg.get("value_start", np.min(radar.azimuth["data"]))
            value_stop = dscfg.get("value_stop", np.max(radar.azimuth["data"]))
            ang_tol = dscfg.get("AngTol", 1.0)
            rng_tol = dscfg.get("RngTol", 50.0)
            fixed_range = dscfg.get("fixed_range", 0.0)
            fixed_azimuth = dscfg.get("fixed_azimuth", None)
            fixed_elevation = dscfg.get("fixed_elevation", 0.0)
        elif mode == "ALONG_ELE":
            value_start = dscfg.get("value_start", np.min(radar.elevation["data"]))
            value_stop = dscfg.get("value_stop", np.max(radar.elevation["data"]))
            ang_tol = dscfg.get("AngTol", 1.0)
            rng_tol = dscfg.get("RngTol", 50.0)
            fixed_range = dscfg.get("fixed_range", 0.0)
            fixed_azimuth = dscfg.get("fixed_azimuth", 0.0)
            fixed_elevation = dscfg.get("fixed_elevation", None)
        else:
            warn("Unknown plotting mode " + dscfg["mode"])
            return None, None

        # initialize dataset
        if not dscfg["initialized"]:
            acoord = pyart.retrieve.compute_ts_along_coord(
                radar,
                field_names,
                mode=mode,
                fixed_range=fixed_range,
                fixed_azimuth=fixed_azimuth,
                fixed_elevation=fixed_elevation,
                ang_tol=ang_tol,
                rng_tol=rng_tol,
                value_start=value_start,
                value_stop=value_stop,
                ref_time=dscfg["timeinfo"],
                acoord=None,
            )

            if acoord is None:
                warn("Unable to compute time series along coordinate")
                return None, None

            global_dict = dict()
            global_dict.update({"start_time": dscfg["timeinfo"]})
            global_dict.update({"radar_out": acoord})
            dscfg["global_data"] = global_dict
            dscfg["initialized"] = 1
        else:
            acoord = pyart.retrieve.compute_ts_along_coord(
                radar,
                field_names,
                mode=mode,
                fixed_range=fixed_range,
                fixed_azimuth=fixed_azimuth,
                fixed_elevation=fixed_elevation,
                ang_tol=ang_tol,
                rng_tol=rng_tol,
                value_start=value_start,
                value_stop=value_stop,
                ref_time=dscfg["timeinfo"],
                acoord=dscfg["global_data"]["radar_out"],
            )

            if acoord is None:
                warn("Unable to compute time series along coordinate")
                return None, None

        dscfg["global_data"]["radar_out"] = acoord

        new_dataset = dict()
        new_dataset.update({"radar_out": acoord})
        new_dataset.update({"radar_type": "temporal"})
        new_dataset.update({"start_time": dscfg["global_data"]["start_time"]})

        return new_dataset, ind_rad

    if procstatus == 2:
        if not dscfg["initialized"]:
            return None, None

        for datatypedescr in dscfg["datatype"]:
            radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
            break

        ind_rad = int(radarnr[5:8]) - 1

        acoord = dscfg["global_data"]["radar_out"]

        new_dataset = dict()
        new_dataset.update({"radar_out": acoord})
        new_dataset.update({"radar_type": "final"})
        new_dataset.update({"start_time": dscfg["global_data"]["start_time"]})

        return new_dataset, ind_rad



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions to retrieve MeteoSwiss products from the archives

Daniel Wolfensberger, Rebecca Gugerli
MeteoSwiss/EPFL
daniel.wolfensberger@epfl.ch, rebecca.gugerli@epfl.ch
December 2019, July 2022

NOTE: these codes are very specific to MeteoSwiss and have no
practical use to external people
"""

import numpy as np
import os
import zipfile
import datetime
import glob
import socket
import subprocess
import fnmatch
import re
import pandas as pd  # function used in retrieve_hzt_prod
from io import BytesIO

# MeteoSwiss constants
OFFSET_CCS4 = [297, -100]
# Folder depends on server:
if ("lom" in socket.gethostname()) or ("meteoswiss" in socket.gethostname()):
    FOLDER_RADAR = "/srn/data/"
    FOLDER_ISO0 = "/srn/data/HZT/"
elif "balfrin" in socket.gethostname() or "nid" in socket.gethostname():
    FOLDER_DATABASE = "/store_new/mch/msrad/radar/radar_database/"
    FOLDER_RADAR = "/store_new/mch//msrad/radar/swiss/data/"
    FOLDER_RADARH = "/store_new/mch/msrad/radar/polarHR/data/"
    FOLDER_CPCCV = "/store_new/mch/msrad/radar/cpc_validation/daily/"
    FOLDER_ISO0 = "/store_new/mch/msrad/radar/swiss/data/"
    FOLDER_RADAR_HDF5 = "/store_new/mch/msrad/radar/swiss/data/hdf5/"


def _make_timezone_aware(dt, tz=datetime.timezone.utc):
    """
    Makes a naive datetime timezone-aware by setting it to the provided timezone.
    If the datetime is already timezone-aware, it is returned unchanged.

    Parameters:
    dt (datetime): The datetime object to check.
    tz (timezone): The timezone to set if the datetime is naive (default is UTC).

    Returns:
    datetime: The timezone-aware datetime object.
    """
    if dt.tzinfo is None:
        return dt.replace(tzinfo=tz)
    return dt


def retrieve_hzt_prod(folder_out, start_time, end_time, pattern_type="shell"):
    """Retrieves the preprocessed HZT products from the CSCS repository for a specified
    time range, unzips them and places them in a specified folder

    Parameters
    ----------

    folder_out: str
        directory where to store the unzipped files
    start_time : datetime.datetime instance
        starting time of the time range
    end_time : datetime.datetime instance
        end time of the time range
    pattern_type: either 'shell' or 'regex' (optional)
        use 'shell' for standard shell patterns, which use * as wildcard
        use 'regex' for more advanced regex patterns

    Returns
    -------
    A list containing all the filepaths of the retrieved files

    """
    dt = datetime.timedelta(hours=1)
    delta = end_time - start_time
    if delta.total_seconds() == 0:
        times = [start_time]
    else:
        times = start_time + np.arange(int(delta.total_seconds() / (60 * 60)) + 2) * dt
    dates = []
    for t in times:
        dates.append(datetime.datetime(year=t.year, month=t.month, day=t.day))
    dates = np.unique(dates)

    t0 = _make_timezone_aware(start_time)
    t1 = _make_timezone_aware(end_time)

    all_files = []
    for i, d in enumerate(dates):
        if i == 0:
            start_time = datetime.datetime(
                year=t0.year, month=t0.month, day=t0.day, hour=t0.hour
            )
            # print('*first start time: ', start_time)
        else:
            start_time = datetime.datetime(year=d.year, month=d.month, day=d.day)
            # print('*all other start times', start_time)
        if i == len(dates) - 1:
            end_time = datetime.datetime(
                year=t1.year, month=t1.month, day=t1.day, hour=t1.hour
            ) + datetime.timedelta(hours=1)
        else:
            end_time = datetime.datetime(year=d.year, month=d.month, day=d.day, hour=23)
            # print('*end_time: ', end_time)

        files = _retrieve_hzt_prod_daily(folder_out, start_time, end_time, pattern_type)

        if files is not None:
            all_files.extend(files)

    return all_files


def retrieve_hzt_RT(tstep):
    """Retrieves the preprocessed HZT products
        A version adapted to real time implementation
        Only used in for the function retrieve_hzt_prod

    Parameters
    ----------

    tstep: datetime
        directory where to store the unzipped files

    Returns
    -------
    A list containing all the filepaths of the retrieved files

    """

    # Get list of available files
    folder_in = FOLDER_ISO0
    content_zip = np.array(
        [
            c
            for c in os.listdir(folder_in)
            if (len(c.split(".")) == 2) and (int(c.split(".")[-1]) >= 800)
        ]
    )

    # HZT files are produced once an hour
    start_time = tstep.replace(minute=0)
    end_time = start_time + datetime.timedelta(hours=1)
    start_time = _make_timezone_aware(start_time)
    end_time = _make_timezone_aware(end_time)

    try:
        # Sort filelist to most recent prediction
        content_filt = np.array([c for c in content_zip if c.endswith("800")])
        times_filt = np.array(
            [
                datetime.datetime.strptime(c[3:12], "%y%j%H%M").replace(
                    tzinfo=datetime.timezone.utc
                )
                + datetime.timedelta(hours=int(c[-2::]))
                for c in content_filt
            ]
        )
        conditions = np.array(
            [np.logical_and((t >= start_time), (t <= end_time)) for t in times_filt]
        )

        content_filt = content_filt[conditions]
        times_filt = times_filt[conditions]
    except (ValueError, TypeError, IndexError):
        print("HZT data does not exist for " + start_time.strftime("%d-%b-%y"))
        files = None
        return

    # Check that an hourly estimate is available
    all_hours = pd.date_range(start=start_time, end=end_time, freq="H")

    if len(all_hours) != len(times_filt):
        content_times = np.array(
            [
                datetime.datetime.strptime(c[3:12], "%y%j%H%M").replace(
                    tzinfo=datetime.timezone.utc
                )
                + datetime.timedelta(hours=int(c[-2::]))
                for c in content_zip
            ]
        )
        # Find time that is missing:
        for hh in all_hours:
            if hh not in times_filt:
                hh_last = np.where(hh == content_times)
                times_filt = np.sort(np.append(times_filt, content_times[hh_last][-1]))
                content_filt = np.sort(
                    np.append(content_filt, content_zip[hh_last][-1])
                )

    # Get a list of all files to retrieve
    conditions = np.array(
        [np.logical_and(t >= start_time, t <= end_time) for t in times_filt]
    )

    if not np.any(conditions):
        msg = """
        No file was found corresponding to this format, verify pattern and product_name
        """
        raise ValueError(msg)

    files = sorted(
        np.array([folder_in + c for c in np.array(content_filt)[conditions]])
    )

    return files


def _retrieve_hzt_prod_daily(folder_out, start_time, end_time, pattern_type="shell"):
    """Retrieves the preprocessed HZT products from the CSCS repository for a day,
        Only used in for the function retrieve_hzt_prod

    Parameters
    ----------

    folder_out: str
        directory where to store the unzipped files
    start_time : datetime.datetime instance
        starting time of the time range
    end_time : datetime.datetime instance
        end time of the time range
    pattern_type: either 'shell' or 'regex' (optional)
        use 'shell' for standard shell patterns, which use * as wildcard
        use 'regex' for more advanced regex patterns

    Returns
    -------
    A list containing all the filepaths of the retrieved files

    """

    folder_out += "/"
    start_time = _make_timezone_aware(start_time)
    end_time = _make_timezone_aware(end_time)

    suffix = str(start_time.year)[-2:] + str(start_time.timetuple().tm_yday).zfill(3)
    folder_in = FOLDER_ISO0 + str(start_time.year) + "/" + suffix + "/"
    name_zipfile = "HZT" + suffix + ".zip"

    try:
        # Get list of files in zipfile
        zipp = zipfile.ZipFile(folder_in + name_zipfile)
        content_zip = np.sort(np.array(zipp.namelist()))

        # Sort filelist to most recent prediction
        content_filt = np.array([c for c in content_zip if c.endswith("800")])
        times_filt = np.array(
            [
                datetime.datetime.strptime(c[3:12], "%y%j%H%M")
                + datetime.timedelta(hours=int(c[-2::]))
                for c in content_filt
            ]
        )
        content_filt = content_filt[
            np.where((times_filt >= start_time) & (times_filt <= end_time))
        ]
        times_filt = times_filt[
            np.where((times_filt >= start_time) & (times_filt <= end_time))
        ]
    except (ValueError, TypeError, IndexError):
        print(
            "Zip file with HZT data does not exist for "
            + start_time.strftime("%d-%b-%y")
        )
        files = None
        return

    # Check that an hourly estimate is available
    all_hours = pd.date_range(start=start_time, end=end_time, freq="H")

    if len(all_hours) != len(times_filt):
        content_times = np.array(
            [
                datetime.datetime.strptime(c[3:12], "%y%j%H%M")
                + datetime.timedelta(hours=int(c[-2::]))
                for c in content_zip
            ]
        )
        # Find time that is missing:
        for hh in all_hours:
            if hh not in times_filt:
                hh_last = np.where(hh == content_times)
                times_filt = np.sort(np.append(times_filt, content_times[hh_last][-1]))
                content_filt = np.sort(
                    np.append(content_filt, content_zip[hh_last][-1])
                )

    # Get a list of all files to retrieve
    conditions = np.array(
        [np.logical_and(t >= start_time, t <= end_time) for t in times_filt]
    )

    if not np.any(conditions):
        msg = """
        No file was found corresponding to this format, verify pattern and product_name
        """
        raise ValueError(msg)

    files_to_retrieve = " ".join(content_filt[conditions])

    # Check if files are already unzipped (saves time if they already exist)
    for fi in content_filt[conditions]:
        if os.path.exists(folder_out + fi):
            files_to_retrieve = files_to_retrieve.replace(fi, "")

    # Only unzip if at least one file does not exist
    if len(files_to_retrieve.strip()) > 0:
        print("Unzippping: " + files_to_retrieve)
        cmd = 'unzip -j -o -qq "{:s}" {:s} -d {:s}'.format(
            folder_in + name_zipfile, files_to_retrieve, folder_out
        )
        subprocess.call(cmd, shell=True)

    files = sorted(np.array([folder_out + c for c in content_filt[conditions]]))

    return files


def retrieve_mch_prod(
    start_time,
    end_time,
    product_name,
    folder_out=None,
    pattern=None,
    pattern_type="shell",
    sweeps=None,
    hdf5=False,
):
    """Retrieves radar data from the CSCS repository for a specified
    time range, unzips them and places them in a specified folder

    Parameters
    ----------

    start_time : datetime.datetime instance
        starting time of the time range
    end_time : datetime.datetime instance
        end time of the time range
    product_name: str
        name of the product, as stored on CSCS, e.g. RZC, CPCH, MZC, BZC...
    folder_out: str
        directory where to store the unzipped files, if set to None
        will read the file to memory
    pattern: str
        pattern constraint on file names, can be used for products which contain
        multiple filetypes, f.ex CPCH folders contain both rda and gif files,
        if only gifs are wanted : file_type = '*.gif'
    pattern_type: either 'shell' or 'regex' (optional)
        use 'shell' for standard shell patterns, which use * as wildcard
        use 'regex' for more advanced regex patterns
    sweeps: list of int (optional)
        For polar products, specifies which sweeps (elevations) must be
        retrieved, if not specified all available sweeps will be retrieved
    hdf5: bool
        If True will retrieve the hdf5 files for the given product (beware
        hdf5 is not available for all products)

    Returns
    -------
    A list containing all the filepaths of the retrieved files

    """

    if product_name == "ZZW" or product_name == "ZZP":  # no vpr for PPM and WEI
        product_name = "ZZA"

    if folder_out:
        if not os.path.exists(folder_out):
            os.makedirs(folder_out)
        if product_name == "CPC":
            folder_out = os.path.join(folder_out, "CPC")
        if product_name == "CPCH":
            folder_out = os.path.join(folder_out, "CPCH")

    # Check if times are aware or naive
    if start_time.tzinfo is None:
        start_time = start_time.replace(tzinfo=datetime.timezone.utc)
    if end_time.tzinfo is None:
        end_time = end_time.replace(tzinfo=datetime.timezone.utc)

    dt = datetime.timedelta(minutes=5)
    delta = end_time - start_time
    if delta.total_seconds() == 0:
        times = [start_time]
    else:
        times = start_time + np.arange(int(delta.total_seconds() / (5 * 60)) + 1) * dt
    dates = []
    for t in times:
        dates.append(datetime.datetime(year=t.year, month=t.month, day=t.day))
    dates = np.unique(dates)

    t0 = start_time
    t1 = end_time

    all_files = []
    for i, d in enumerate(dates):
        if i == 0:
            start_time = t0
        else:
            start_time = datetime.datetime(
                year=d.year, month=d.month, day=d.day, tzinfo=datetime.timezone.utc
            )
        if i == len(dates) - 1:
            end_time = t1
        else:
            end_time = datetime.datetime(
                year=d.year,
                month=d.month,
                day=d.day,
                hour=23,
                minute=59,
                tzinfo=datetime.timezone.utc,
            )
        files = _retrieve_prod_daily(
            start_time,
            end_time,
            product_name,
            folder_out,
            pattern,
            pattern_type,
            sweeps,
            hdf5,
        )

        all_files.extend(files)

    return all_files


def retrieve_mch_prod_RT(
    time, product_name, pattern=None, pattern_type="shell", sweeps=None
):
    """Adapted function from rainforest.common.retrieve_data
        Here, it reads the data per timestep, and in the real-time
        operation, the radar data is not zipped

    Args:
        time (datetime object): timestamp to extract
        product_name (string): Name of the product to be extracted
        sweeps (list): List of sweeps if not all want to be extracted. Defaults to None.

    Raises:
        ValueError: If no data is found

    Returns:
        dict: dictionary containing with the the file list
    """
    time = _make_timezone_aware(time)

    # Get all files
    folder_radar = FOLDER_RADAR
    folder_in = folder_radar + product_name + "/"

    # Get list of available files
    content_zip = np.array(os.listdir(folder_in))

    if pattern is not None:
        if pattern_type == "shell":
            content_zip = [
                c for c in content_zip if fnmatch.fnmatch(os.path.basename(c), pattern)
            ]
        elif pattern_type == "regex":
            content_zip = [
                c
                for c in content_zip
                if re.match(pattern, os.path.basename(c)) is not None
            ]
        else:
            raise ValueError('Unknown pattern_type, must be either "shell" or "regex".')

    # Derive datetime of each file
    times_zip = np.array(
        [
            datetime.datetime.strptime(c[3:12], "%y%j%H%M").replace(
                tzinfo=datetime.timezone.utc
            )
            for c in content_zip
        ]
    )

    # Get a list of all files to retrieve
    conditions = times_zip == time

    # Filter on sweeps:
    if sweeps is not None:
        sweeps_zip = np.array([int(c[-3:]) for c in content_zip])
        # Get a list of all files to retrieve
        conditions_sweep = np.array([s in sweeps for s in sweeps_zip])
        conditions = np.logical_and(conditions, conditions_sweep)

    if not np.any(conditions):
        msg = """
        No file was found corresponding to this format, verify pattern and product_name
        """
        raise ValueError(msg)

    files = sorted(np.array([folder_in + c for c in np.array(content_zip)[conditions]]))

    return files


def _retrieve_prod_daily(
    start_time,
    end_time,
    product_name,
    folder_out=None,
    pattern=None,
    pattern_type="shell",
    sweeps=None,
    hdf5=False,
):
    """Retrieve radar product files for a given day, with an option to store them in RAM."""

    if hdf5:
        folder_radar = FOLDER_RADAR_HDF5
    else:
        folder_radar = FOLDER_RADARH if product_name[:2] == "MH" else FOLDER_RADAR

    suffix = str(start_time.year)[-2:] + str(start_time.timetuple().tm_yday).zfill(3)
    folder_in = os.path.join(folder_radar, str(start_time.year), suffix)
    name_zipfile = product_name + suffix + ".zip"

    # Open the zip file
    with zipfile.ZipFile(os.path.join(folder_in, name_zipfile), "r") as zipp:
        content_zip = np.array(zipp.namelist())

        # Filter files based on pattern if provided
        if pattern:
            if pattern_type == "shell":
                content_zip = [
                    c
                    for c in content_zip
                    if fnmatch.fnmatch(os.path.basename(c), pattern)
                ]
            elif pattern_type == "regex":
                content_zip = [
                    c for c in content_zip if re.match(pattern, os.path.basename(c))
                ]
            else:
                raise ValueError(
                    'Unknown pattern_type, must be either "shell" or "regex".'
                )

        content_zip = np.array(content_zip)

        # Extract timestamps from filenames
        times_zip = np.array(
            [
                datetime.datetime.strptime(c[3:12], "%y%j%H%M").replace(
                    tzinfo=datetime.timezone.utc
                )
                for c in content_zip
            ]
        )

        # Filter files based on the given time range
        conditions = np.array([start_time <= t <= end_time for t in times_zip])

        # Further filter based on sweeps if provided
        if sweeps is not None:
            sweeps_zip = np.array([int(c[-3:]) for c in content_zip])
            conditions_sweep = np.array([s in sweeps for s in sweeps_zip])
            conditions = np.logical_and(conditions, conditions_sweep)

        if not np.any(conditions):
            raise ValueError(
                "No file was found corresponding to this format, verify pattern and product_name"
            )

        selected_files = content_zip[conditions]
        if not folder_out:
            # Load selected files into memory as dictionary {filename: file_content}
            files_in_memory = [
                BytesIO(zipp.read(file_name)) for file_name in selected_files
            ]
            return files_in_memory
        else:
            # Prepare files for extraction
            files_to_retrieve = " ".join(selected_files)

            # Check if files are already unzipped (skip those that exist)
            for fi in selected_files:
                if os.path.exists(folder_out + fi):
                    files_to_retrieve = files_to_retrieve.replace(fi, "")

            # Unzip only if needed
            if len(files_to_retrieve.strip()) > 0:
                cmd = f'unzip -j -o -qq "{os.path.join(folder_in, name_zipfile)}" {files_to_retrieve} -d {folder_out}'
                subprocess.call(cmd, shell=True)

            files = sorted([folder_out + c for c in selected_files])
            return files


def retrieve_CPCCV(time, stations):
    """Retrieves cross-validation CPC data for a set of stations from
    the xls files prepared by Yanni

    Parameters
    ----------

    time : datetime.datetime instance
        starting time of the time range
    stations : list of str
        list of weather stations at which to retrieve the CPC.CV data

    Returns
    -------
    A numpy array corresponding at the CPC.CV estimations at every specified
    station
    """

    time = _make_timezone_aware(time)

    from ..io.read_data_other import read_xls

    year = time.year

    folder = FOLDER_CPCCV + str(year) + "/"

    files = sorted([f for f in glob.glob(folder + "*.xls") if ".s" not in f])

    def _start_time(fname):
        bname = os.path.basename(fname)
        times = bname.split(".")[1]
        tend = times.split("_")[1]
        return datetime.datetime.strptime(tend, "%Y%m%d%H00")

    tend = np.array([_start_time(f) for f in files])

    match = np.where(time < tend)[0]

    if not len(match):
        print("Could not find CPC CV file for time {:s}".format(time))
        return np.zeros((len(stations))) + np.nan

    data = read_xls(files[match[0]])

    hour = int(datetime.datetime.strftime(time, "%Y%m%d%H00"))
    idx = np.where(np.array(data["time.stamp"]) == hour)[0]
    data_hour = data.iloc[idx]
    data_hour_stations = data_hour.iloc[
        np.isin(np.array(data_hour["nat.abbr"]), stations)
    ]
    cpc_cv = []
    cpc_xls = []
    for sta in stations:
        if sta in np.array(data_hour_stations["nat.abbr"]):
            cpc_cv.append(
                float(
                    data_hour_stations.loc[data_hour_stations["nat.abbr"] == sta][
                        "CPC.CV"
                    ]
                )
            )
            cpc_xls.append(
                float(
                    data_hour_stations.loc[data_hour_stations["nat.abbr"] == sta]["CPC"]
                )
            )
        else:
            cpc_cv.append(np.nan)
            cpc_xls.append(np.nan)

    return np.array(cpc_cv), np.array(cpc_xls)


def retrieve_AQC_XLS(time, stations):
    """Retrieves cross-validation CPC data for a set of stations from
    the xls files prepared by Yanni

    Parameters
    ----------

    time : datetime.datetime instance
        starting time of the time range
    stations : list of str
        list of weather stations at which to retrieve the CPC.CV data

    Returns
    -------
    A numpy array corresponding at the CPC.CV estimations at every specified
    station
    """
    from ..io.read_data_other import read_xls

    time = _make_timezone_aware(time)
    year = time.year

    folder = FOLDER_CPCCV + str(year) + "/"

    files = sorted([f for f in glob.glob(folder + "*.xls") if ".s" not in f])

    def _start_time(fname):
        bname = os.path.basename(fname)
        times = bname.split(".")[1]
        tend = times.split("_")[1]
        return datetime.datetime.strptime(tend, "%Y%m%d%H00")

    tend = np.array([_start_time(f) for f in files])

    match = np.where(time < tend)[0]

    if not len(match):
        print("Could not find CPC CV file for time {:s}".format(time))
        return np.zeros((len(stations))) + np.nan

    data = read_xls(files[match[0]])

    hour = int(datetime.datetime.strftime(time, "%Y%m%d%H00"))
    idx = np.where(np.array(data["time.stamp"]) == hour)[0]
    data_hour = data.iloc[idx]
    data_hour_stations = data_hour.iloc[
        np.isin(np.array(data_hour["nat.abbr"]), stations)
    ]
    aqc_xls = []
    for sta in stations:
        if sta in np.array(data_hour_stations["nat.abbr"]):
            aqc_xls.append(
                float(
                    data_hour_stations.loc[data_hour_stations["nat.abbr"] == sta]["AQC"]
                )
            )
        else:
            aqc_xls.append(np.nan)

    return np.array(aqc_xls)



try:
    import fcntl

    FCNTL_AVAIL = True
except ModuleNotFoundError:
    import msvcrt  # For windows

    FCNTL_AVAIL = False

import os
import time
from warnings import warn
import errno


def lock_file(file):
    if FCNTL_AVAIL:
        while True:
            try:
                fcntl.flock(file, fcntl.LOCK_EX | fcntl.LOCK_NB)
                break
            except OSError as e:
                if e.errno == errno.EAGAIN:
                    time.sleep(0.1)
                elif e.errno == errno.EBADF:
                    warn(
                        "WARNING: No file locking is possible (NFS mount?), "
                        + "expect strange issues with multiprocessing..."
                    )
                    break
                else:
                    raise
    else:
        while True:
            try:
                # Attempt to acquire an exclusive lock on the file
                msvcrt.locking(
                    os.open(file, os.O_RDWR | os.O_CREAT), msvcrt.LK_NBLCK, 0
                )
                break
            except OSError as e:
                if e.errno == 13:  # Permission denied
                    time.sleep(0.1)
                elif e.errno == 13:  # No such file or directory
                    warn(
                        "WARNING: No file locking is possible (NFS mount?), "
                        + "expect strange issues with multiprocessing..."
                    )
                    break
                else:
                    raise


def unlock_file(file):
    if os.name == "posix":
        fcntl.flock(file, fcntl.LOCK_UN)
    else:
        msvcrt.locking(file.fileno(), msvcrt.LK_UNLCK, 1)



"""
==================================
Input and output (:mod:`pyrad.io`)
==================================

.. currentmodule:: pyrad.io

Functions to read and write data and configuration files.

Reading configuration files
===========================

.. autosummary::
    :toctree: generated/

    read_config

Reading radar data
==================

.. autosummary::
    :toctree: generated/

    get_data

Reading icon data
==================

.. autosummary::
    :toctree: generated/

    icon2radar_data
    icon2radar_coord
    hzt2radar_data
    hzt2radar_coord
    get_icon_fields
    get_iso0_field
    read_icon_data
    read_icon_coord
    read_hzt_data
    read_iso0_mf_data
    read_iso0_grib_data
    iso2radar_data
    grib2radar_data
    get_iso0_ref

Reading DEM data
==================

.. autosummary::
    :toctree: generated/

    dem2radar_data
    read_idrisi_data
    read_idrisi_metadata

Reading other data
==================

.. autosummary::
    :toctree: generated/

    read_centroids_npz
    read_centroids
    read_proc_periods
    read_last_state
    read_status
    read_rad4alp_icon
    read_rad4alp_vis
    read_excess_gates
    read_colocated_gates
    read_colocated_data
    read_timeseries
    read_ts_cum
    read_monitoring_ts
    read_intercomp_scores_ts
    get_sensor_data
    read_smn
    read_smn2
    read_coord_sensors
    read_disdro_scattering
    read_disdro_parsivel
    read_knmi
    read_sun_hits
    read_sun_hits_multiple_days
    read_sun_retrieval
    read_solar_flux
    read_selfconsistency
    read_antenna_pattern
    read_meteorage
    read_lightning
    read_lightning_traj
    read_lightning_all
    read_trt_scores
    read_trt_data
    read_trt_traj_data
    read_trt_thundertracking_traj_data
    read_trt_cell_lightning
    read_trt_info_all
    read_trt_info_all2
    read_trt_info
    read_trt_info2
    read_thundertracking_info
    read_rhi_profile
    read_vpr_theo_parameters
    read_histogram
    read_quantiles
    read_profile_ts
    read_histogram_ts
    read_quantiles_ts
    read_ml_ts
    read_windmills_data
    read_radiosounding_wyoming
    read_radiosounding_igra
    read_fzl_igra
    read_meteoswiss_xml_vad

Writing data
==================

.. autosummary::
    :toctree: generated/

    write_vol_csv
    write_vol_kml
    write_centroids
    write_proc_periods
    write_ts_lightning
    send_msg
    write_alarm_msg
    write_last_state
    write_smn
    write_trt_info
    write_trt_thundertracking_data
    write_trt_cell_data
    write_trt_cell_scores
    write_trt_cell_lightning
    write_trt_rpc
    write_vpr_theo_params
    write_vpr_info
    write_rhi_profile
    write_field_coverage
    write_cdf
    write_histogram
    write_quantiles
    write_multiple_points
    write_multiple_points_grid
    write_ts_polar_data
    write_ts_grid_data
    write_ts_cum
    write_ts_stats
    write_monitoring_ts
    write_excess_gates
    write_intercomp_scores_ts
    write_colocated_gates
    write_colocated_data
    write_colocated_data_time_avg
    write_sun_hits
    write_sun_retrieval
    write_fixed_angle


Auxiliary functions
===================

.. autosummary::
    :toctree: generated/

    get_rad4alp_prod_fname
    get_datatype_skyecho
    map_hydro
    map_Doppler
    get_save_dir
    make_filename
    generate_field_name_str
    get_fieldname_pyart
    get_fieldname_icon
    get_field_unit
    get_file_list
    get_file_list_s3
    get_scan_files_to_merge
    get_scan_files_to_merge_s3
    get_rad4alp_dir
    get_rad4alp_grid_dir
    get_trtfile_list
    get_new_rainbow_file_name
    get_datatype_fields
    get_dataset_fields
    get_datetime
    find_raw_icon_file
    find_hzt_file
    find_iso0_file
    find_iso0_grib_file
    find_date_in_file_name
    _get_datetime

Trajectory
==========

.. autosummary::
    :toctree: generated/

    Trajectory

TimeSeries
==========

.. autosummary::
    :toctree: generated/

    TimeSeries

"""

from .config import read_config  # noqa

from .read_data_radar import get_data, add_field, interpol_field  # noqa

from .read_data_icon import read_icon_data, read_icon_coord  # noqa
from .read_data_icon import icon2radar_data, icon2radar_coord  # noqa
from .read_data_icon import get_icon_fields  # noqa

from .read_data_dem import read_idrisi_data, read_idrisi_metadata  # noqa
from .read_data_dem import dem2radar_data  # noqa

from .read_data_hzt import read_hzt_data, hzt2radar_data, hzt2radar_coord  # noqa
from .read_data_hzt import get_iso0_field  # noqa

from .read_data_iso0_mf import read_iso0_mf_data, iso2radar_data, get_iso0_ref  # noqa
from .read_data_iso0_mf import read_iso0_grib_data, grib2radar_data  # noqa

from .read_data_other import read_status, read_rad4alp_icon, read_rad4alp_vis  # noqa
from .read_data_other import read_timeseries, read_monitoring_ts, read_ts_cum  # noqa
from .read_data_other import read_intercomp_scores_ts, read_quantiles  # noqa
from .read_data_other import read_selfconsistency, read_colocated_gates  # noqa
from .read_data_other import read_colocated_data, read_antenna_pattern  # noqa
from .read_data_other import read_colocated_data_time_avg  # noqa
from .read_data_other import read_last_state, read_rhi_profile, read_centroids  # noqa
from .read_data_other import read_excess_gates, read_histogram  # noqa
from .read_data_other import read_profile_ts, read_histogram_ts  # noqa
from .read_data_other import read_quantiles_ts, read_ml_ts, read_proc_periods  # noqa
from .read_data_other import read_centroids_npz, read_mf_vis  # noqa
from .read_data_other import read_vpr_theo_parameters  # noqa
from .read_data_other import read_mch_xml_vad  # noqa

from .read_data_sensor import read_lightning, read_lightning_traj  # noqa
from .read_data_sensor import get_sensor_data, read_smn, read_smn2  # noqa
from .read_data_sensor import read_disdro_scattering, read_trt_data  # noqa
from .read_data_sensor import read_trt_traj_data, read_lightning_all  # noqa
from .read_data_sensor import read_trt_scores, read_trt_cell_lightning  # noqa
from .read_data_sensor import read_meteorage, read_trt_info_all, read_trt_info  # noqa
from .read_data_sensor import read_thundertracking_info, read_windmills_data  # noqa
from .read_data_sensor import read_trt_info2, read_trt_info_all2  # noqa
from .read_data_sensor import read_trt_thundertracking_traj_data  # noqa
from .read_data_sensor import read_coord_sensors, read_disdro_parsivel  # noqa
from .read_data_sensor import read_knmi  # noqa
from .read_data_sensor import (  # noqa
    read_radiosounding_wyoming,
    read_radiosounding_igra,
)
from .read_data_sensor import read_fzl_igra  # noqa

from .read_data_sun import read_sun_hits_multiple_days, read_sun_hits  # noqa
from .read_data_sun import read_sun_retrieval, read_solar_flux  # noqa

from .write_data import write_smn, write_ts_polar_data, write_ts_cum  # noqa
from .write_data import write_monitoring_ts, write_intercomp_scores_ts  # noqa
from .write_data import write_sun_hits, write_sun_retrieval  # noqa
from .write_data import write_colocated_gates, write_colocated_data  # noqa
from .write_data import write_colocated_data_time_avg, write_cdf  # noqa
from .write_data import write_rhi_profile, write_field_coverage  # noqa
from .write_data import write_last_state, write_alarm_msg, send_msg  # noqa
from .write_data import write_excess_gates, write_trt_cell_data  # noqa
from .write_data import write_histogram, write_quantiles, write_ts_lightning  # noqa
from .write_data import write_trt_cell_scores, write_trt_cell_lightning  # noqa
from .write_data import write_trt_info, write_fixed_angle, write_proc_periods  # noqa
from .write_data import write_trt_thundertracking_data, write_ts_grid_data  # noqa
from .write_data import write_trt_rpc, write_ts_stats, write_centroids  # noqa
from .write_data import write_multiple_points, write_vpr_theo_params  # noqa
from .write_data import write_multiple_points_grid, write_vpr_info  # noqa
from .write_data import write_vol_kml, write_vol_csv  # noqa

from .io_aux import get_save_dir, make_filename, get_new_rainbow_file_name  # noqa
from .io_aux import get_datetime, get_dataset_fields, map_hydro, map_Doppler  # noqa
from .io_aux import get_file_list, get_trtfile_list, get_datatype_fields  # noqa
from .io_aux import get_fieldname_pyart, get_field_unit, get_fieldname_icon  # noqa
from .io_aux import generate_field_name_str, find_raw_icon_file  # noqa
from .io_aux import find_hzt_file, _get_datetime, get_rad4alp_prod_fname  # noqa
from .io_aux import get_rad4alp_dir, get_rad4alp_grid_dir, find_iso0_file  # noqa
from .io_aux import find_iso0_grib_file, find_date_in_file_name  # noqa
from .io_aux import get_datatype_skyecho, get_file_list_s3  # noqa
from .io_aux import get_scan_files_to_merge, get_scan_files_to_merge_s3  # noqa

from .trajectory import Trajectory  # noqa

from .timeseries import TimeSeries  # noqa

__all__ = [s for s in dir() if not s.startswith("_")]



"""
pyrad.flow.flow_control
=======================

functions to control the Pyrad data processing flow

.. autosummary::
    :toctree: generated/

    main
    main_rt
    main_cosmo
    main_cosmo_rt
    main_gecsx
"""

from __future__ import print_function
import warnings
from warnings import warn
import traceback
import os
from datetime import datetime
from datetime import timedelta
import gc
import subprocess
import queue
import time
from pathlib import Path

from pyart import __version__ as pyart_version
from .. import version as pyrad_version

from .flow_aux import _warning_format, _initialize_listener
from .flow_aux import _create_cfg_dict, _create_datacfg_dict
from .flow_aux import _get_times_and_traj, _get_datatype_list
from .flow_aux import _get_datasets_list, _get_masterfile_list
from .flow_aux import _wait_for_files, _get_radars_data
from .flow_aux import _initialize_datasets
from .flow_aux import _process_datasets, _postprocess_datasets

from ..io.io_aux import get_datetime
from ..io.read_data_other import read_last_state
from ..io.write_data import write_last_state

ALLOW_USER_BREAK = False

try:
    import dask
    from dask.diagnostics import Profiler, ResourceProfiler, CacheProfiler
    from dask.diagnostics import visualize
    from distributed import Client
    from bokeh.io import export_png

    _DASK_AVAILABLE = True
except ImportError:
    warn("dask not available: The processing will not be parallelized")
    _DASK_AVAILABLE = False


def main(
    cfgfile,
    starttime=None,
    endtime=None,
    trajfile="",
    trajtype="plane",
    flashnr=0,
    infostr="",
    MULTIPROCESSING_DSET=False,
    MULTIPROCESSING_PROD=False,
    PROFILE_MULTIPROCESSING=False,
    USE_CHILD_PROCESS=False,
):
    """
    Main flow control. Processes radar data off-line over a period of time
    given either by the user, a trajectory file, or determined by the last
    volume processed and the current time. Multiple radars can be processed
    simultaneously

    Parameters
    ----------
    cfgfile : str
        path of the main config file
    starttime, endtime : datetime object
        start and end time of the data to be processed
    trajfile : str
        path to file describing the trajectory
    trajtype : str
        type of trajectory file. Can be either 'plane', 'lightning' or
        'proc_periods'
    flashnr : int
        If larger than 0 will select a flash in a lightning trajectory file.
        If 0 the data corresponding to the trajectory of all flashes will be
        plotted
    infostr : str
        Information string about the actual data processing
        (e.g. 'RUN57'). This string is added to product files.
    MULTIPROCESSING_DSET : Bool
        If true the generation of datasets at the same processing level will
        be parallelized
    MULTIPROCESSING_PROD : Bool
        If true the generation of products from each dataset will be
        parallelized
    PROFILE_MULTIPROCESSING : Bool
        If true and code parallelized the multiprocessing is profiled
    USE_CHILD_PROCESS : Bool
        If true the reading and processing of the data will be performed by
        a child process controlled by dask. This is done to make sure all
        memory used is released.

    """
    print(
        "- PYRAD version: {} (compiled {} by {})".format(
            pyrad_version.version,
            pyrad_version.compile_date_time,
            pyrad_version.username,
        )
    )
    print("- PYART version: {}".format(pyart_version))

    # Define behaviour of warnings
    warnings.simplefilter("always")  # always print matching warnings
    # warnings.simplefilter('error')  # turn matching warnings into exceptions
    warnings.formatwarning = _warning_format  # define format

    if ALLOW_USER_BREAK:
        input_queue = _initialize_listener()

    if not _DASK_AVAILABLE:
        MULTIPROCESSING_DSET = False
        MULTIPROCESSING_PROD = False
        PROFILE_MULTIPROCESSING = False
        USE_CHILD_PROCESS = False

    # check if multiprocessing profiling is necessary
    if not MULTIPROCESSING_DSET and not MULTIPROCESSING_PROD and not USE_CHILD_PROCESS:
        PROFILE_MULTIPROCESSING = False
    elif (
        int(MULTIPROCESSING_DSET) + int(MULTIPROCESSING_PROD) + int(USE_CHILD_PROCESS)
        > 1
    ):
        PROFILE_MULTIPROCESSING = False

    if (
        int(MULTIPROCESSING_DSET) + int(MULTIPROCESSING_PROD) + int(USE_CHILD_PROCESS)
        > 1
    ):
        # necessary to launch tasks from tasks
        Client()

    if PROFILE_MULTIPROCESSING:
        prof = Profiler()
        rprof = ResourceProfiler()
        cprof = CacheProfiler()

        prof.register()
        rprof.register()
        cprof.register()

    cfg = _create_cfg_dict(cfgfile)
    datacfg = _create_datacfg_dict(cfg)

    starttimes, endtimes, traj = _get_times_and_traj(
        trajfile,
        starttime,
        endtime,
        cfg["ScanPeriod"],
        last_state_file=cfg["lastStateFile"],
        trajtype=trajtype,
        flashnr=flashnr,
    )

    if infostr:
        print("- Info string : {}".format(infostr))

    # get data types and levels
    datatypesdescr_list = list()
    for i in range(1, cfg["NumRadars"] + 1):
        datatypesdescr_list.append(
            _get_datatype_list(cfg, radarnr="RADAR" + "{:03d}".format(i))
        )
    dataset_levels = _get_datasets_list(cfg)

    masterfilelist, masterdatatypedescr, masterscan = _get_masterfile_list(
        datatypesdescr_list[0],
        starttimes,
        endtimes,
        datacfg,
        scan_list=datacfg["ScanList"],
    )

    nvolumes = len(masterfilelist)
    if nvolumes == 0:
        raise ValueError(
            "ERROR: Could not find any valid volumes between "
            "{} and {} for master scan '{}' and master data type '{}'".format(
                starttimes[0].strftime("%Y-%m-%d %H:%M:%S"),
                endtimes[-1].strftime("%Y-%m-%d %H:%M:%S"),
                masterscan,
                masterdatatypedescr,
            )
        )
    print("- Number of volumes to process: {}".format(nvolumes))
    print("- Start time: {}".format(starttimes[0].strftime("%Y-%m-%d %H:%M:%S")))
    print("- end time: {}".format(endtimes[-1].strftime("%Y-%m-%d %H:%M:%S")))

    # initial processing of the datasets
    print("\n\n- Initializing datasets:")
    dscfg, traj = _initialize_datasets(dataset_levels, cfg, traj=traj, infostr=infostr)

    # process all data files in file list or until user interrupts processing
    for masterfile in masterfilelist:
        if ALLOW_USER_BREAK:
            # check if user has requested exit
            try:
                input_queue.get_nowait()
                warn("Program terminated by user")
                break
            except queue.Empty:
                pass

        print("\n- master file: {}".format(os.path.basename(masterfile)))

        master_voltime = get_datetime(masterfile, masterdatatypedescr)

        if USE_CHILD_PROCESS:
            data_reading = dask.delayed(_get_radars_data)(
                master_voltime,
                datatypesdescr_list,
                datacfg,
                num_radars=datacfg["NumRadars"],
            )

            try:
                radar_list = data_reading.compute()
                del data_reading

                dscfg_aux = dask.delayed(dscfg)
                traj_aux = dask.delayed(traj)
                data_processing = dask.delayed(_process_datasets)(
                    dataset_levels,
                    cfg,
                    dscfg_aux,
                    radar_list,
                    master_voltime,
                    traj=traj_aux,
                    infostr=infostr,
                    MULTIPROCESSING_DSET=MULTIPROCESSING_DSET,
                    MULTIPROCESSING_PROD=MULTIPROCESSING_PROD,
                )
                try:
                    dscfg, traj = data_processing.compute()
                    del data_processing
                    del radar_list
                    del dscfg_aux
                    del traj_aux

                except Exception as ee:
                    warn(str(ee))
                    traceback.print_exc()
            except Exception as ee:
                warn(str(ee))
                traceback.print_exc()

        else:
            radar_list = _get_radars_data(
                master_voltime,
                datatypesdescr_list,
                datacfg,
                num_radars=datacfg["NumRadars"],
            )

            # process all data sets
            dscfg, traj = _process_datasets(
                dataset_levels,
                cfg,
                dscfg,
                radar_list,
                master_voltime,
                traj=traj,
                infostr=infostr,
                MULTIPROCESSING_DSET=MULTIPROCESSING_DSET,
                MULTIPROCESSING_PROD=MULTIPROCESSING_PROD,
            )

            # delete variables
            del radar_list

        gc.collect()

    # post-processing of the datasets
    print("\n\n- Post-processing datasets:")
    dscfg, traj = _postprocess_datasets(
        dataset_levels, cfg, dscfg, traj=traj, infostr=infostr
    )

    if PROFILE_MULTIPROCESSING:
        prof.unregister()
        rprof.unregister()
        cprof.unregister()

        bokeh_plot = visualize([prof, rprof, cprof], show=False, save=False)

        profile_path = os.path.expanduser("~") + "/profiling/"
        if not os.path.isdir(profile_path):
            os.makedirs(profile_path)

        export_png(
            bokeh_plot,
            filename=(
                profile_path
                + datetime.utcnow().strftime("%Y%m%d%H%M%S")
                + "_profile.png"
            ),
        )

    print("- This is the end my friend! See you soon!")


def main_rt(
    cfgfile_list,
    starttime=None,
    endtime=None,
    infostr_list=None,
    proc_period=60,
    proc_finish=None,
    hide_warnings=False,
):
    """
    main flow control. Processes radar data in real time. The start and end
    processing times can be determined by the user. This function is inteded
    for a single radar

    Parameters
    ----------
    cfgfile_list : list of str
        path of the main config files
    starttime, endtime : datetime object
        start and end time of the data to be processed
    infostr_list : list of str
        Information string about the actual data processing
        (e.g. 'RUN57'). This string is added to product files.
    proc_period : int
        period of time before starting a new processing round (seconds)
    proc_finish : int or None
        if set to a value the program will be forced to shut down after the
        value (in seconds) from start time has been exceeded
    hide_warnings : bool
        if set to true it will hide the warnings during the pyrad processing
        this is useful when logging the outputs to log files, as they can
        become very large

    Returns
    -------
    end_proc : Boolean
        If true the program has ended successfully

    """
    print(
        "- PYRAD version: %s (compiled %s by %s)"
        % (
            pyrad_version.version,
            pyrad_version.compile_date_time,
            pyrad_version.username,
        )
    )
    print("- PYART version: " + pyart_version)

    if hide_warnings:
        warnings.filterwarnings("ignore")
    else:
        # Define behaviour of warnings
        warnings.simplefilter("always")  # always print matching warnings
        # warnings.simplefilter('error')  # turn matching warnings into
        # exceptions
        warnings.formatwarning = _warning_format  # define format

    # The processing will be allowed to run for a limited period
    if proc_finish is not None:
        startime_proc = datetime.utcnow()
        # for offline testing
        # startime_proc = startime_proc.replace(
        #     year=endtime.year, month=endtime.month, day=endtime.day)
        # startime_proc = startime_proc.replace(hour=10)

        endtime_proc = startime_proc + timedelta(seconds=proc_finish)

    if ALLOW_USER_BREAK:
        input_queue = _initialize_listener()

    cfg_list = []
    datacfg_list = []
    dscfg_list = []
    datatypesdescr_list_list = []
    dataset_levels_list = []
    last_processed_list = []

    for icfg, cfgfile in enumerate(cfgfile_list):
        cfg = _create_cfg_dict(cfgfile)
        if infostr_list is not None:
            infostr = infostr_list[icfg]
        else:
            infostr = ""
        datacfg = _create_datacfg_dict(cfg)

        if infostr:
            print("- Info string : " + infostr)

        # find out last processed volume
        last_processed = read_last_state(cfg["lastStateFile"])
        if last_processed is None:
            print("- last processed volume unknown")
        else:
            print("- last processed volume: " + last_processed.strftime("%Y%m%d%H%M%S"))
        last_processed_list.append(last_processed)

        # get data types and levels
        datatypesdescr_list = list()
        for i in range(1, cfg["NumRadars"] + 1):
            datatypesdescr_list.append(
                _get_datatype_list(cfg, radarnr="RADAR" + "{:03d}".format(i))
            )

        dataset_levels = _get_datasets_list(cfg)

        # initial processing of the datasets
        print("\n\n- Initializing datasets:")
        dscfg, traj = _initialize_datasets(dataset_levels, cfg, infostr=infostr)

        cfg_list.append(cfg)
        datacfg_list.append(datacfg)
        dscfg_list.append(dscfg)
        datatypesdescr_list_list.append(datatypesdescr_list)
        dataset_levels_list.append(dataset_levels)

        # remove variables from memory
        del cfg
        del datacfg
        del dscfg
        del datatypesdescr_list
        del dataset_levels
        del last_processed
        del traj

        gc.collect()

    end_proc = False
    while not end_proc:
        if ALLOW_USER_BREAK:
            # check if user has requested exit
            try:
                user_input = input_queue.get_nowait()
                end_proc = user_input
                warn("Program terminated by user")
                break
            except queue.Empty:
                pass

        nowtime = datetime.utcnow()
        # for offline testing
        # nowtime = nowtime.replace(
        #     year=endtime.year, month=endtime.month, day=endtime.day)
        # nowtime = nowtime.replace(hour=10)

        # if processing end time exceeded finalize processing
        if proc_finish is not None:
            if nowtime >= endtime_proc:
                end_proc = True
                warn("Allowed processing time exceeded")
                break

        # end time has been set and current time older than end time
        # quit processing
        if endtime is not None:
            if nowtime > endtime:
                end_proc = True
                break

        # start time has been set. Check if current time has to be
        # processed. If not sleep until next proc_period
        if starttime is not None:
            if nowtime < starttime:
                time.sleep(proc_period)
                continue

        vol_processed = False
        for icfg, cfg in enumerate(cfg_list):
            if ALLOW_USER_BREAK:
                # check if user has requested exit
                try:
                    user_input = input_queue.get_nowait()
                    end_proc = user_input
                    warn("Program terminated by user")
                    break
                except queue.Empty:
                    pass

            datacfg = datacfg_list[icfg]
            dscfg = dscfg_list[icfg]
            datatypesdescr_list = datatypesdescr_list_list[icfg]
            dataset_levels = dataset_levels_list[icfg]
            last_processed = last_processed_list[icfg]
            if infostr_list is not None:
                infostr = infostr_list[icfg]
            else:
                infostr = ""

            # wait until new files are available
            masterfile, masterdatatypedescr, last_processed = _wait_for_files(
                nowtime, datacfg, datatypesdescr_list[0], last_processed=last_processed
            )
            if masterfile is None:
                last_processed_list[icfg] = last_processed
                if last_processed is not None:
                    write_last_state(last_processed, cfg["lastStateFile"])
                continue

            print("\n- master file: " + os.path.basename(masterfile))
            master_voltime = get_datetime(masterfile, masterdatatypedescr)

            # get data of master radar
            radar_list = _get_radars_data(
                master_voltime,
                datatypesdescr_list,
                datacfg,
                num_radars=datacfg["NumRadars"],
            )

            # process all data sets
            dscfg, traj = _process_datasets(
                dataset_levels, cfg, dscfg, radar_list, master_voltime, infostr=infostr
            )

            last_processed_list[icfg] = master_voltime
            write_last_state(master_voltime, cfg["lastStateFile"])
            dscfg_list[icfg] = dscfg

            vol_processed = True

            # remove variables from memory
            del radar_list
            del cfg
            del datacfg
            del dscfg
            del datatypesdescr_list
            del dataset_levels
            del last_processed
            del traj

            gc.collect()

        nowtime_new = datetime.utcnow()
        # for offline testing
        # nowtime_new = nowtime_new.replace(
        #     year=endtime.year, month=endtime.month, day=endtime.day)
        # nowtime_new = nowtime_new.replace(hour=10)

        proc_time = (nowtime_new - nowtime).total_seconds()
        if vol_processed:
            print("Processing time %s s\n" % proc_time)

        # if processing end time exceeded finalize processing
        if proc_finish is not None:
            if nowtime_new >= endtime_proc:
                end_proc = True
                warn("Allowed processing time exceeded")
                break

        if proc_time < proc_period:
            time.sleep(proc_period - proc_time)

    # only do post processing if program properly terminated by user
    if end_proc:
        # post-processing of the datasets
        print("\n\n- Post-processing datasets:")
        for icfg, cfg in enumerate(cfg_list):
            dscfg = dscfg_list[icfg]
            dataset_levels = dataset_levels_list[icfg]
            if infostr_list is not None:
                infostr = infostr_list[icfg]
            else:
                infostr = ""
            dscfg, traj = _postprocess_datasets(
                dataset_levels, cfg, dscfg, infostr=None
            )

            # remove variables from memory
            del cfg
            del dscfg
            del dataset_levels
            del traj

            gc.collect()

    print("- This is the end my friend! See you soon!")

    return end_proc


def main_cosmo(cfgfile, starttime=None, endtime=None, trajfile="", infostr=""):
    """
    Main flow control. Processes radar data off-line over a period of time
    given either by the user, a trajectory file, or determined by the last
    volume processed and the current time. Multiple radars can be processed
    simultaneously

    Parameters
    ----------
    cfgfile : str
        path of the main config file
    starttime, endtime : datetime object
        start and end time of the data to be processed
    trajfile : str
        path to file describing the trajectory
    infostr : str
        Information string about the actual data processing
        (e.g. 'RUN57'). This string is added to product files.

    """
    print(
        "- PYRAD version: {} (compiled {} by {})".format(
            pyrad_version.version,
            pyrad_version.compile_date_time,
            pyrad_version.username,
        )
    )
    print("- PYART version: {}".format(pyart_version))

    # Define behaviour of warnings
    warnings.simplefilter("always")  # always print matching warnings
    # warnings.simplefilter('error')  # turn matching warnings into exceptions
    warnings.formatwarning = _warning_format  # define format

    cfg = _create_cfg_dict(cfgfile)
    datacfg = _create_datacfg_dict(cfg)

    starttimes, endtimes, traj = _get_times_and_traj(
        trajfile,
        starttime,
        endtime,
        cfg["ScanPeriod"],
        last_state_file=cfg["lastStateFile"],
        trajtype="proc_periods",
    )

    if infostr:
        print("- Info string : {}".format(infostr))

    # get data types and levels
    datatypesdescr_list = list()
    for i in range(1, cfg["NumRadars"] + 1):
        datatypesdescr_list.append(
            _get_datatype_list(cfg, radarnr="RADAR" + "{:03d}".format(i))
        )

    dataset_levels = _get_datasets_list(cfg)

    masterfilelist, masterdatatypedescr, masterscan = _get_masterfile_list(
        datatypesdescr_list[0], starttimes, endtimes, datacfg
    )

    nvolumes = len(masterfilelist)
    if nvolumes == 0:
        raise ValueError(
            "ERROR: Could not find any valid COSMO data between "
            "{} and {}".format(
                starttimes[0].strftime("%Y-%m-%d %H:%M:%S"),
                endtimes[-1].strftime("%Y-%m-%d %H:%M:%S"),
            )
        )
    print("- Number of volumes to process: {}".format(nvolumes))
    print("- Start time: {}".format(starttimes[0].strftime("%Y-%m-%d %H:%M:%S")))
    print("- end time: {}".format(endtimes[-1].strftime("%Y-%m-%d %H:%M:%S")))

    # initial processing of the datasets
    print("\n\n- Initializing datasets:")
    dscfg, traj = _initialize_datasets(dataset_levels, cfg, traj=traj, infostr=infostr)

    # process all data files in file list
    for masterfile in masterfilelist:
        print("\n- master file: {}".format(os.path.basename(masterfile)))

        master_voltime = get_datetime(masterfile, masterdatatypedescr)

        # process all data sets
        dscfg, traj = _process_datasets(
            dataset_levels, cfg, dscfg, None, master_voltime, traj=traj, infostr=infostr
        )

        gc.collect()

    print("- This is the end my friend! See you soon!")


def main_cosmo_rt(
    cfgfile_list,
    starttime=None,
    endtime=None,
    infostr_list=None,
    proc_period=60,
    proc_finish=None,
):
    """
    main flow control. Processes radar data in real time. The start and end
    processing times can be determined by the user. This function is inteded
    for a single radar

    Parameters
    ----------
    cfgfile_list : list of str
        path of the main config files
    starttime, endtime : datetime object
        start and end time of the data to be processed
    infostr_list : list of str
        Information string about the actual data processing
        (e.g. 'RUN57'). This string is added to product files.
    proc_period : int
        period of time before starting a new processing round (seconds)
    proc_finish : int or None
        if set to a value the program will be forced to shut down after the
        value (in seconds) from start time has been exceeded

    Returns
    -------
    end_proc : Boolean
        If true the program has ended successfully

    """
    print(
        "- PYRAD version: %s (compiled %s by %s)"
        % (
            pyrad_version.version,
            pyrad_version.compile_date_time,
            pyrad_version.username,
        )
    )
    print("- PYART version: " + pyart_version)

    # Define behaviour of warnings
    warnings.simplefilter("always")  # always print matching warnings
    # warnings.simplefilter('error')  # turn matching warnings into exceptions
    warnings.formatwarning = _warning_format  # define format

    # The processing will be allowed to run for a limited period
    if proc_finish is not None:
        startime_proc = datetime.utcnow()
        # for offline testing
        # startime_proc = startime_proc.replace(
        #     year=endtime.year, month=endtime.month, day=endtime.day)
        # startime_proc = startime_proc.replace(hour=10)

        endtime_proc = startime_proc + timedelta(seconds=proc_finish)

    if ALLOW_USER_BREAK:
        input_queue = _initialize_listener()

    cfg_list = []
    datacfg_list = []
    dscfg_list = []
    datatypesdescr_list_list = []
    dataset_levels_list = []
    last_processed_list = []

    for icfg, cfgfile in enumerate(cfgfile_list):
        cfg = _create_cfg_dict(cfgfile)
        if infostr_list is not None:
            infostr = infostr_list[icfg]
        else:
            infostr = ""
        datacfg = _create_datacfg_dict(cfg)

        if infostr:
            print("- Info string : " + infostr)

        # find out last processed volume
        last_processed = read_last_state(cfg["lastStateFile"])
        if last_processed is None:
            print("- last processed volume unknown")
        else:
            print("- last processed volume: " + last_processed.strftime("%Y%m%d%H%M%S"))
        last_processed_list.append(last_processed)

        # get data types and levels
        datatypesdescr_list = list()
        for i in range(1, cfg["NumRadars"] + 1):
            datatypesdescr_list.append(
                _get_datatype_list(cfg, radarnr="RADAR" + "{:03d}".format(i))
            )

        dataset_levels = _get_datasets_list(cfg)

        # initial processing of the datasets
        print("\n\n- Initializing datasets:")
        dscfg, traj = _initialize_datasets(dataset_levels, cfg, infostr=infostr)

        cfg_list.append(cfg)
        datacfg_list.append(datacfg)
        dscfg_list.append(dscfg)
        datatypesdescr_list_list.append(datatypesdescr_list)
        dataset_levels_list.append(dataset_levels)

        # remove variables from memory
        del cfg
        del datacfg
        del dscfg
        del datatypesdescr_list
        del dataset_levels
        del last_processed
        del traj

        gc.collect()

    end_proc = False
    while not end_proc:
        if ALLOW_USER_BREAK:
            # check if user has requested exit
            try:
                user_input = input_queue.get_nowait()
                end_proc = user_input
                warn("Program terminated by user")
                break
            except queue.Empty:
                pass

        nowtime = datetime.utcnow()
        # for offline testing
        # nowtime = nowtime.replace(
        #     year=endtime.year, month=endtime.month, day=endtime.day)
        # nowtime = nowtime.replace(hour=10)

        # if processing end time exceeded finalize processing
        if proc_finish is not None:
            if nowtime >= endtime_proc:
                end_proc = True
                warn("Allowed processing time exceeded")
                break

        # end time has been set and current time older than end time
        # quit processing
        if endtime is not None:
            if nowtime > endtime:
                end_proc = True
                break

        # start time has been set. Check if current time has to be
        # processed. If not sleep until next proc_period
        if starttime is not None:
            if nowtime < starttime:
                time.sleep(proc_period)
                continue

        vol_processed = False
        for icfg, cfg in enumerate(cfg_list):
            if ALLOW_USER_BREAK:
                # check if user has requested exit
                try:
                    user_input = input_queue.get_nowait()
                    end_proc = user_input
                    warn("Program terminated by user")
                    break
                except queue.Empty:
                    pass

            datacfg = datacfg_list[icfg]
            dscfg = dscfg_list[icfg]
            datatypesdescr_list = datatypesdescr_list_list[icfg]
            dataset_levels = dataset_levels_list[icfg]
            last_processed = last_processed_list[icfg]
            if infostr_list is not None:
                infostr = infostr_list[icfg]
            else:
                infostr = ""

            # wait until new files are available
            masterfile, masterdatatypedescr, last_processed = _wait_for_files(
                nowtime, datacfg, datatypesdescr_list[0], last_processed=last_processed
            )
            if masterfile is None:
                last_processed_list[icfg] = last_processed
                if last_processed is not None:
                    write_last_state(last_processed, cfg["lastStateFile"])
                continue

            print("\n- master file: " + os.path.basename(masterfile))
            master_voltime = get_datetime(masterfile, masterdatatypedescr)

            # process all data sets
            dscfg, traj = _process_datasets(
                dataset_levels, cfg, dscfg, None, master_voltime, infostr=infostr
            )

            last_processed_list[icfg] = master_voltime
            write_last_state(master_voltime, cfg["lastStateFile"])
            dscfg_list[icfg] = dscfg

            vol_processed = True

            # remove variables from memory
            del cfg
            del datacfg
            del dscfg
            del datatypesdescr_list
            del dataset_levels
            del last_processed
            del traj

            gc.collect()

        nowtime_new = datetime.utcnow()
        # for offline testing
        # nowtime_new = nowtime_new.replace(
        #     year=endtime.year, month=endtime.month, day=endtime.day)
        # nowtime_new = nowtime_new.replace(hour=10)

        proc_time = (nowtime_new - nowtime).total_seconds()
        if vol_processed:
            print("Processing time %s s\n" % proc_time)

        # if processing end time exceeded finalize processing
        if proc_finish is not None:
            if nowtime_new >= endtime_proc:
                end_proc = True
                warn("Allowed processing time exceeded")
                break

        if proc_time < proc_period:
            time.sleep(proc_period - proc_time)

    print("- This is the end my friend! See you soon!")

    return end_proc


def main_gecsx(cfgfile, starttime=None, endtime=None, infostr="", gather_plots=True):
    """
    Main flow control. Processes radar data off-line over a period of time
    given either by the user, a trajectory file, or determined by the last
    volume processed and the current time. Multiple radars can be processed
    simultaneously

    Parameters
    ----------
    cfgfile : str
        path of the main config file
    time, endtime : datetime object
        start and end time of the radar data to be used as reference for
        gecsx
    infostr : str
        Information string about the actual data processing
        (e.g. 'RUN57'). This string is added to product files.
    """
    GECSX_MANDATORY = [
        "frequency",
        "radar_beam_width_h",
        "pulse_width",
        "txpwrh",
        "AntennaGainH",
        "mflossh",
        "lrxh",
    ]

    GECSX_OPTIONAL = ["attg", "AzimTol", "mosotti_factor", "refcorr", "RadarPosition"]

    print(
        "- PYRAD version: {} (compiled {} by {})".format(
            pyrad_version.version,
            pyrad_version.compile_date_time,
            pyrad_version.username,
        )
    )
    print("- PYART version: {}".format(pyart_version))

    # Define behaviour of warnings
    warnings.simplefilter("always")  # always print matching warnings
    # warnings.simplefilter('error')  # turn matching warnings into exceptions
    warnings.formatwarning = _warning_format  # define format

    cfg = _create_cfg_dict(cfgfile)
    datacfg = _create_datacfg_dict(cfg)

    if starttime is None:
        starttime = datetime.now()
        endtime = starttime
    if endtime is None:
        endtime = datetime.now()

    if infostr:
        print("- Info string : {}".format(infostr))

    if cfg["datapath"] is not None:
        starttimes, endtimes, _ = _get_times_and_traj(
            None,
            starttime,
            endtime,
            cfg["ScanPeriod"],
            last_state_file=cfg["lastStateFile"],
            trajtype="proc_periods",
        )

    # get data types and levels
    datatypesdescr_list = list()
    for i in range(1, cfg["NumRadars"] + 1):
        datatypesdescr_list.append(
            _get_datatype_list(cfg, radarnr="RADAR" + "{:03d}".format(i))
        )

    dataset_levels = _get_datasets_list(cfg)

    if cfg["datapath"] is not None:
        masterfilelist, masterdatatypedescr, _ = _get_masterfile_list(
            datatypesdescr_list[0], starttimes, endtimes, datacfg, datacfg["ScanList"]
        )

        nvolumes = len(masterfilelist)
        if nvolumes == 0:
            warn(
                f"WARNING: Could not find any valid radar data between "
                f"{starttimes[0].strftime('%Y-%m-%d %H:%M:%S')} and "
                f"{endtimes[-1].strftime('%Y-%m-%d %H:%M:%S')}"
            )
            warn("WARNING: Proceeding without radar reference")
        print("- Number of volumes to process: {}".format(nvolumes))
        print("- Start time: {}".format(starttimes[0].strftime("%Y-%m-%d %H:%M:%S")))
        print("- end time: {}".format(endtimes[-1].strftime("%Y-%m-%d %H:%M:%S")))
    else:
        masterfilelist = []
    # For GECSX we treat only one master file, since the method does not
    # depend on time
    if len(masterfilelist) != 0:
        masterfile = masterfilelist[0]
        master_voltime = get_datetime(masterfile, masterdatatypedescr)

        # get data of master radar
        radar_list = _get_radars_data(master_voltime, datatypesdescr_list, datacfg)
    else:
        radar_list = []

    # initial processing of the datasets
    print("\n\n- Initializing datasets:")
    dscfg, _ = _initialize_datasets(dataset_levels, cfg, infostr=infostr)

    # For GECSX we copy some keys from the datacfg dict to the dataset dict
    for dset in dscfg.values():
        for k in GECSX_MANDATORY:
            if k in dset.keys():
                continue
            try:
                dset[k] = datacfg[k]
            except Exception:
                msg = f"Mandatory GECSX key {k} is missing from loc file"
                raise ValueError(msg)
        for k in GECSX_OPTIONAL:
            if k in datacfg.keys():
                dset[k] = datacfg[k]

        if len(radar_list) == 0:
            valid_radarpos = False
            if "RadarPosition" in dset:
                if (
                    "latitude" in dset["RadarPosition"]
                    and "longitude" in dset["RadarPosition"]
                    and "altitude" in dset["RadarPosition"]
                ):
                    valid_radarpos = True
            if not valid_radarpos:
                raise ValueError(
                    "When no radar data is provided, the structure "
                    '"RadarPosition" with field "altitude", "latitude" '
                    'and "longitude" must be provided in the loc file'
                )
            for k in dset["RadarPosition"].keys():
                if not isinstance(dset["RadarPosition"][k], list):
                    dset["RadarPosition"][k] = [dset["RadarPosition"][k]]

    # process all data sets
    dscfg, _ = _process_datasets(
        dataset_levels, cfg, dscfg, radar_list, None, None, infostr=infostr
    )

    gc.collect()

    if gather_plots:
        img_ext = cfg["imgformat"]
        for dset in dscfg:
            gather_dir = str(Path(cfg["saveimgbasepath"], cfg["name"], dset))
            print(
                "Copying all generated figures into dir {:s}...".format(
                    gather_dir + "/ALL_FIGURES/"
                )
            )
            for ex in img_ext:
                if os.path.exists(gather_dir):
                    cmd = (
                        'cd {:s}; mkdir -p ALL_FIGURES; find . -type f -name "*.{:s}" '.format(
                            gather_dir, ex
                        )
                        + "-exec cp {} ALL_FIGURES \\;"
                    )
                    subprocess.call(cmd, shell=True)

    print("- This is the end my friend! See you soon!")



"""
pyrad.io.read_data_dem
========================

Functions for reading data derived from Digital Elevation Models (DEM)

.. autosummary::
    :toctree: generated/

    dem2radar_data
    read_idrisi_data
    read_idrisi_metadata
    _prepare_for_interpolation


"""

import pathlib
from warnings import warn
import numpy as np
from scipy.interpolate import RegularGridInterpolator

# check existence of gdal
try:
    from osgeo import gdal, osr

    _GDAL_AVAILABLE = True
except ImportError:
    try:
        import gdal
        import osr

        _GDAL_AVAILABLE = True
    except ImportError:
        _GDAL_AVAILABLE = False

import pyart
from pyart.config import get_metadata
from .read_data_icon import _put_radar_in_swiss_coord

# from memory_profiler import profile

# import time


def dem2radar_data(radar, dem_data, slice_xy=True, field_name="visibility"):
    """
    get the DEM value corresponding to each radar gate using nearest
    neighbour interpolation

    Parameters
    ----------
    radar : Radar
        the radar object containing the information on the position of the
        radar gates
    dem_data : dict
        dictionary containing the DEM data
    slice_xy : boolean
        if true the horizontal plane of the DEM field is cut to the
        dimensions of the radar field
    field_names : str
        names of DEM fields to convert

    Returns
    -------
    dem_field : dict
        Dictionary with the DEM fields and metadata

    """
    # debugging
    # start_time = time.time()

    x_radar, y_radar, _ = _put_radar_in_swiss_coord(radar)

    (x_dem, y_dem, ind_xmin, ind_ymin, ind_xmax, ind_ymax) = _prepare_for_interpolation(
        x_radar, y_radar, dem_data, slice_xy=slice_xy
    )

    if field_name not in dem_data:
        warn("DEM field " + field_name + " data not available")
        return None

    values = dem_data[field_name]["data"][
        ind_xmin : ind_xmax + 1, ind_ymin : ind_ymax + 1
    ]

    # Note RegularGridInterpolator is 10x faster than NDNearestInterpolator
    # and has the advantage of not extrapolating outside of grid domain

    # replace masked values with nans
    values = np.ma.filled(values, np.nan)
    interp_func = RegularGridInterpolator((x_dem, y_dem), values, bounds_error=False)

    # interpolate
    data_interp = interp_func((x_radar, y_radar))

    del values
    # restore mask
    data_interp = np.ma.masked_equal(data_interp, np.nan)

    # put field
    field_dict = get_metadata(field_name)
    field_dict["data"] = data_interp.astype(float)

    del data_interp

    return field_dict


# @profile


def read_dem(fname, field_name="terrain_altitude", fill_value=None, projparams=None):
    """
    Generic reader that reads DEM data from any format, will infer the proper
    reader from filename extension

    Parameters
    ----------
    fname : str
        name of the file to read
    field_name : str
        name of the readed variable
    fill_value : float
        The fill value, if not provided will be infered from metadata
        if possible
    projparams : str or int
        projection transform as can be used by gdal either a  Proj4
        string, see epsg.io for a list, or a EPSG number, if not provided
        will be infered from the file, or for ASCII, LV1903 will be used

    Returns
    -------
    dem_data : dictionary
        dictionary with the data and metadata

    """
    extension = pathlib.Path(fname).suffix

    if isinstance(projparams, int):  # Retrieve Wkt code from EPSG number
        proj = osr.SpatialReference()
        proj.ImportFromEPSG(projparams)
        projparams = proj.ExportToProj4()
        projparams = _proj4_str_to_dict(projparams)

    if extension in [".tif", ".tiff", ".gtif"]:
        dem = read_geotiff_data(fname, fill_value)
    elif extension in [".asc", ".dem", ".txt"]:
        dem = read_ascii_data(fname, fill_value)
    elif extension in [".rst"]:
        dem = read_idrisi_data(fname, fill_value)
    else:
        warn(
            "WARNING: unable to read file %s, extension must be .tif .tiff .gtif, "
            + ".asc .dem .txt .rst".format()
        )
        return None

    if "projparams" in dem:
        projparams = dem["projparams"]

    field_dict = get_metadata(field_name)
    field_dict["data"] = dem["data"][::-1, :][None, :, :]
    field_dict["units"] = dem["metadata"]["value units"]

    x = get_metadata("x")
    y = get_metadata("y")
    z = get_metadata("z")

    orig_lat = get_metadata("origin_latitude")
    orig_lon = get_metadata("origin_longitude")
    orig_alt = get_metadata("origin_altitude")

    x["data"] = (
        np.arange(dem["metadata"]["columns"]) * dem["metadata"]["resolution"]
        + dem["metadata"]["resolution"] / 2.0
        + dem["metadata"]["min. X"]
    )

    y["data"] = (
        np.arange(dem["metadata"]["rows"]) * dem["metadata"]["resolution"]
        + dem["metadata"]["resolution"] / 2.0
        + dem["metadata"]["min. Y"]
    )

    z["data"] = np.array([0])

    orig_lat["data"] = [y["data"][0]]
    orig_lon["data"] = [x["data"][0]]
    orig_alt["data"] = [0]

    if projparams is None:
        warn(
            "WARNING: No proj could be read from file and no projparams were provided, "
            + "assuming the projection is CH1903."
        )
        projparams = _get_lv1903_proj4()

    time = get_metadata("grid_time")
    time["data"] = np.array([0.0])
    time["units"] = "seconds since 2000-01-01T00:00:00Z"

    # The use of CRS().to_dict() is required to use GridMapDisplay of Pyart
    # which expects a dict for the projection attribute of the grid
    dem_data = pyart.core.Grid(
        time,
        {field_name: field_dict},
        dem["metadata"],
        orig_lat,
        orig_lon,
        orig_alt,
        x,
        y,
        z,
        projection=projparams,
    )

    return dem_data


# @profile


def read_geotiff_data(fname, fill_value=None):
    """
    Reads DEM data from a generic geotiff file, unit system is expected to be
    in meters!

    Parameters
    ----------
    fname : str
        name of the file to read
    field_name : str
        name of the readed variable
    fill_value : float
        The fill value, if not provided will be infered from metadata
        (recommended)
    projparams : projection transform as can be used by pyproj, either a
        EPSG code integer (21781 for CH1903 for example), or a OGC WKT or
        Proj4 string, see epsg.io for a list,
        if not provided will be infered from the idrisi file

    Returns
    -------
    dem : dict
        A dictionary with the following keys:
        - data : array
            data within the DEM
        - metadata : dict
            a dictionary containing the metadata
        - projparams : str
            projection parameters in proj4 format
    """

    if not _GDAL_AVAILABLE:
        warn("gdal is required to use read_geotiff_data but is not installed")
        return None

    # read the data
    try:
        raster = gdal.Open(fname)

        prj = raster.GetProjection()
        try:
            srs = osr.SpatialReference(wkt=prj)
            projparams = _proj4_str_to_dict(srs.ExportToProj4())
        except RuntimeError:
            projparams = None
            pass

        if not projparams:  # Case of empty dict
            projparams = None

        width = raster.RasterXSize
        height = raster.RasterYSize
        gt = raster.GetGeoTransform()
        minx = gt[0]
        miny = gt[3] + width * gt[4] + height * gt[5]

        metadata = {}
        metadata["resolution"] = np.abs(gt[1])
        metadata["min. X"] = minx
        metadata["min. Y"] = miny
        metadata["rows"] = raster.RasterYSize
        metadata["columns"] = raster.RasterXSize
        metadata["value units"] = "meters"
        metadata["max. X"] = (
            metadata["min. X"] + metadata["resolution"] * metadata["columns"]
        )
        metadata["max. Y"] = (
            metadata["min. Y"] + metadata["resolution"] * metadata["rows"]
        )
        metadata["flag value"] = raster.GetRasterBand(1).GetNoDataValue()

        if not fill_value:
            fill_value = metadata["flag value"]

        rasterarray = raster.ReadAsArray()
        rasterarray = np.ma.masked_equal(rasterarray, fill_value)

        dem = {"projparams": projparams, "metadata": metadata, "data": rasterarray}
        return dem

    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None


# @profile


def read_ascii_data(fname, fill_value=None):
    """
    Reads DEM data from an ASCII file in the Swisstopo format, the
    unit system is expected to be in meters!

    Parameters
    ----------
    fname : str
        name of the file to read
    field_name : str
        name of the readed variable
    fill_value : float
        The fill value, if not provided will be infered from metadata
        (recommended)
    projparams : projection transform as can be used by pyproj, either a
        EPSG code integer (21781 for CH1903 for example), or a OGC WKT or
        Proj4 string, see epsg.io for a list,
        if not provided CH1903 (EPSG:21781) will be used


    Returns
    -------
    dem : dict
        A dictionary with the following keys:
        - data : array
            data within the DEM
        - metadata : dict
            a dictionary containing the metadata
    """

    # read the data
    try:
        with open(fname, "r") as file:
            lines = [file.readline().strip().split() for _ in range(6)]

        # Parse metadata
        metadata = {
            "columns": int(lines[0][1]),
            "rows": int(lines[1][1]),
            "min. X": float(lines[2][1]),
            "min. Y": float(lines[3][1]),
            "resolution": float(lines[4][1]),
            "flag value": float(lines[5][1]),
            "max. X": float(lines[2][1]) + float(lines[4][1]) * int(lines[0][1]),
            "max. Y": float(lines[3][1]) + float(lines[4][1]) * int(lines[1][1]),
            "value units": "m",
        }

        # Set fill value if not provided
        if fill_value is None:
            fill_value = metadata["flag value"]

        # Read raster data, skipping first 6 lines
        raster_array = np.loadtxt(fname, skiprows=6)

        # Mask NODATA values
        raster_array = np.ma.masked_equal(raster_array, fill_value)

        dem = {"metadata": metadata, "data": raster_array}

        return dem

    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None


# @profile


def read_idrisi_data(fname, fill_value=None):
    """
    Reads DEM data from an IDRISI .rst file

    Parameters
    ----------
    fname : str
        name of the file to read
    field_name : str
        name of the readed variable
    fill_value : float
        The fill value
    projparams : projection transform as can be used by pyproj, either a
        EPSG code integer (21781 for CH1903 for example), or a OGC WKT or
        Proj4 string, see epsg.io for a list,
        if not provided will be infered from the idrisi file


    Returns
    -------
    dem : dict
        A dictionary with the following keys:
        - data : array
            data within the DEM
        - metadata : dict
            a dictionary containing the metadata
        - projparams : str
            projection parameters in proj4 format
    """

    if not _GDAL_AVAILABLE:
        warn("gdal is required to use read_idrisi_data but is not installed")
        return None

    # read the data
    try:
        if fill_value is None:
            fill_value = -99.0

        raster = gdal.Open(fname)

        prj = raster.GetProjection()
        try:
            srs = osr.SpatialReference(wkt=prj)
            projparams = _proj4_str_to_dict(srs.ExportToProj4())
        except RuntimeError:
            projparams = None
            pass

        if not projparams:  # Case of empty dict
            projparams = None

        rasterarray = raster.ReadAsArray()
        rasterarray = np.ma.masked_equal(rasterarray, fill_value)

        metadata = read_idrisi_metadata(fname)

        if metadata is None:
            return None

        dem = {"projparams": projparams, "metadata": metadata, "data": rasterarray}
        return dem

    except EnvironmentError as ee:
        warn(str(ee))
        warn("Unable to read file " + fname)
        return None


def read_idrisi_metadata(fname):
    """
    Reads DEM metadata from a IDRISI .rdc file

    Parameters
    ----------
    fname : str
        name of the file to read

    Returns
    -------
    metadata : dictionary
        dictionary with the metadata

    """
    # read the data
    fname_rdc = fname.replace(".rst", ".rdc")

    try:
        metadata = dict()
        with open(fname_rdc, "r", newline="") as txtfile:
            for line in txtfile:
                strs = line.split(":")
                metadata.update(
                    {strs[0].strip(): pyart.aux_io.convert_data(strs[1].strip())}
                )

        return metadata
    except EnvironmentError:
        warn("Unable to read file " + fname_rdc)
        return None


def _prepare_for_interpolation(x_radar, y_radar, dem_coord, slice_xy=True):
    """
    prepares the DEM 2D volume for interpolation:
        1. if set slices the DEM data to the area
    covered by the radar
        2. creates the x, y grid for the interpolation

    Parameters
    ----------
    x_radar, y_radar : arrays
        The Swiss coordinates of the radar
    dem_coord : dict
        dictionary containing the DEM coordinates
    slice_xy : boolean
        if true the horizontal plane of the DEM field is cut to the
        dimensions of the radar field

    Returns
    -------
    x_dem, y_dem : 1D arrays
        arrays containing the flatten swiss coordinates of the DEM data in
        the area of interest
    ind_xmin, ind_ymin, ind_xmax, ind_ymax : ints
        the minimum and maximum indices of each dimension

    """
    nx_dem = len(dem_coord["x"]["data"])
    ny_dem = len(dem_coord["y"]["data"])

    if slice_xy:
        # get the D data within the radar range
        xmin = np.min(x_radar)
        xmax = np.max(x_radar)
        ymin = np.min(y_radar)
        ymax = np.max(y_radar)

        ind_xmin = np.where(dem_coord["x"]["data"] < xmin)[0]
        if ind_xmin.size == 0:
            ind_xmin = 0
        else:
            ind_xmin = ind_xmin[-1]

        ind_xmax = np.where(dem_coord["x"]["data"] > xmax)[0]
        if ind_xmax.size == 0:
            ind_xmax = nx_dem - 1
        else:
            ind_xmax = ind_xmax[0]

        ind_ymin = np.where(dem_coord["y"]["data"] < ymin)[0]
        if ind_ymin.size == 0:
            ind_ymin = 0
        else:
            ind_ymin = ind_ymin[-1]

        ind_ymax = np.where(dem_coord["y"]["data"] > ymax)[0]
        if ind_ymax.size == 0:
            ind_ymax = ny_dem - 1
        else:
            ind_ymax = ind_ymax[0]
    else:
        ind_xmin = 0
        ind_xmax = nx_dem - 1
        ind_ymin = 0
        ind_ymax = ny_dem - 1

    x_dem = dem_coord["x"]["data"][ind_xmin : ind_xmax + 1]
    y_dem = dem_coord["y"]["data"][ind_ymin : ind_ymax + 1]

    # Not used with RegularGridInterpolator
    # nx = ind_xmax-ind_xmin+1
    # ny = ind_ymax-ind_ymin+1
    # x_dem = (
    #     np.broadcast_to(x_dem.reshape(nx, 1), (nx, ny))).flatten()
    # y_dem = (
    #     np.broadcast_to(y_dem.reshape(1, ny), (nx, ny))).flatten()

    return (x_dem, y_dem, ind_xmin, ind_ymin, ind_xmax, ind_ymax)


def _proj4_str_to_dict(proj4str):
    # COnverts proj4 string to dict as can be used by part
    return dict(
        item.split("=")
        for item in proj4str.strip(" ").split("+")
        if len(item.split("=")) == 2
    )


def _get_lv1903_proj4():
    # Gets proj4 dict for Swiss LV1903 coordinates
    if _GDAL_AVAILABLE:
        lv1903 = osr.SpatialReference()
        lv1903.ImportFromEPSG(21781)
        lv1903 = lv1903.ExportToProj4()
        # Convert proj4 string to dict
        lv1903 = _proj4_str_to_dict(lv1903)
        if "no_defs" not in lv1903.keys():
            lv1903["no_defs"] = 0
    else:
        # Hardcoded version
        lv1903 = {
            "proj": "somerc",
            "lat_0": 46.9524055555556,
            "lon_0": 7.43958333333333,
            "k_0": 1,
            "x_0": 600000,
            "y_0": 200000,
            "ellps": "bessel",
            "units": "m",
            "no_defs": 0,
            "type": "crs",
        }

    return lv1903
